"""
Test the CommonData API: slicing, maths operations, equality, all main API functions, and no side effects...
"""
from unittest import TestCase
from nose.tools import assert_raises
from cis.test.util.mock import make_mock_cube, make_regular_2d_ungridded_data
from cis.data_io.gridded_data import make_from_cube
import numpy as np
from numpy.testing.utils import assert_array_almost_equal


class TestAPI(TestCase):
    def setUp(self):
        """
            Create the dummy objects necessary for testing
        """
        self.ug = make_regular_2d_ungridded_data()
        self.ug_1 = make_regular_2d_ungridded_data()
        self.gd = make_from_cube(make_mock_cube())
        self.gd_large = make_from_cube(make_mock_cube(50, 30))

    def test_can_slice_ungridded_data(self):
        res = self.ug[2:5]
        assert_array_almost_equal(res.data, self.ug.data[2:5])
        assert_array_almost_equal(res.lat.data, self.ug.lat.data[2:5])
        # This dataset should still be the same as the alternative one (this checks data and metadata)
        assert self.ug == self.ug_1

    def test_can_slice_single_row_of_ungridded_data(self):
        scalar = self.ug[2]
        assert_array_almost_equal(scalar.data, self.ug.data[2])
        assert_array_almost_equal(scalar.lat.data, self.ug.lat.data[2])
        # This dataset should still be the same as the alternative one (this checks data and metadata)
        assert self.ug == self.ug_1

    def test_can_slice_multidim_ungridded_data(self):
        res = self.ug[2:5, 2]
        assert_array_almost_equal(res.data, self.ug.data[2:5, 2])
        assert_array_almost_equal(res.lat.data, self.ug.lat.data[2:5, 2])
        # This dataset should still be the same as the alternative one (this checks data and metadata)
        assert self.ug == self.ug_1

    # Basic maths operations

    def test_can_add_ungridded_data_objects(self):
        res = self.ug + self.ug_1
        assert_array_almost_equal(res.data, self.ug.data + self.ug_1.data)

    def test_can_add_numpy_arrays_and_ungridded_data_objects(self):
        # Create a new object that is the same size so I can just use it's data array
        data = make_regular_2d_ungridded_data().data
        res = self.ug + data
        assert_array_almost_equal(res.data, self.ug.data + data)

    def test_can_add_scalar_and_ungridded_data_objects(self):
        res = self.ug + 5
        assert_array_almost_equal(res.data, self.ug.data + 5)

    def test_can_perform_in_place_maths(self):
        self.ug -= 5
        # Test against the other ungridded data object as this one has changed!
        assert_array_almost_equal(self.ug.data, self.ug_1.data - 5)

    def test_can_perform_unary_op_on_ungridded_data(self):
        from cis.maths import abs
        self.ug -= 5
        res = abs(self.ug)
        # Test against the other ungridded data object as this one has changed!
        assert_array_almost_equal(res.data, np.abs(self.ug_1.data - 5))

    def test_cannot_add_gridded_and_ungridded_data_objects_of_different_shape(self):
        # Force to have the same units
        self.gd_large.units = self.ug.units
        with assert_raises(ValueError):
            res = self.ug + self.gd_large

    def test_CAN_add_gridded_and_ungridded_data_objects_of_same_shape(self):
        # Force to have the same units
        self.gd.units = self.ug.units
        res = self.ug + self.gd
        assert_array_almost_equal(res.data, self.ug.data + self.gd.data)

    def test_can_do_basic_coord_math(self):
        res_c = self.ug.lat + 5
        assert_array_almost_equal(res_c.data, self.ug.lat.data + 5)

    # API tests (subset, aggregate, collocate, etc)

    def test_basic_gridded_to_ungridded_collocation(self):
        res = self.gd.collocated_onto(self.ug)
        res_1 = self.ug.sampled_from(self.gd)
        # The two results should be identical
        assert res == res_1
        # This dataset should still be the same as the alternative one (this checks data and metadata)
        assert self.ug == self.ug_1

    def test_basic_ungridded_to_ungridded_collocation(self):
        res = self.ug.collocated_onto(self.ug_1)
        res_1 = self.ug_1.sampled_from(self.ug)
        # The two results should be identical
        assert res == res_1
        # This dataset should still be the same as the alternative one (this checks data and metadata)
        assert self.ug == self.ug_1

    def test_basic_gridded_to_gridded_collocation(self):
        gd_copy = self.gd.copy()
        res = self.gd.collocated_onto(self.gd)
        assert self.gd == gd_copy
        res_1 = self.gd.sampled_from(self.gd)
        # The two data results should be identical, although the metadata is slightly different for some reason
        assert_array_almost_equal(res[0].data, res_1[0].data)

    def test_basic_ungridded_to_gridded_collocation(self):
        gd_copy = self.gd.copy()
        res = self.ug.collocated_onto(self.gd)
        assert self.gd == gd_copy
        res_1 = self.gd.sampled_from(self.ug)
        # The two data results should be identical, although the metadata is slightly different for some reason
        assert_array_almost_equal(res[0].data, res_1[0].data)
        # This dataset should still be the same as the alternative one (this checks data and metadata)
        assert self.ug == self.ug_1

    def test_basic_gridded_subsetting(self):
        gd_copy = self.gd.copy()
        res = self.gd.subset(x=[-5, 0])
        assert self.gd == gd_copy

    def test_basic_ungridded_subsetting(self):
        res = self.ug.subset(x=[-5, 0])
        # This dataset should still be the same as the alternative one (this checks data and metadata)
        assert self.ug == self.ug_1

    def test_basic_gridded_collapse(self):
        gd_copy = self.gd.copy()
        res = self.gd.collapsed('x')
        assert self.gd == gd_copy

    def test_basic_gridded_collapse_using_coord_objects(self):
        gd_copy = self.gd.copy()
        res = self.gd.collapsed(self.gd.coords())
        assert self.gd == gd_copy

    def test_basic_ungridded_aggregation(self):
        res = self.ug.aggregate(x=[-5, 0, 1])
        # This dataset should still be the same as the alternative one (this checks data and metadata)
        assert self.ug == self.ug_1

    def test_copy_copies_metadata(self):
        res = self.ug.copy()
        res.var_name = 'test'
        assert self.ug.var_name != 'test'

    def test_change_units(self):
        from numpy.testing import assert_almost_equal
        assert self.ug.data[0, 0] == 1

        self.ug.units = 'cm-3'

        self.ug.convert_units('m-3')

        assert str(self.ug.units) == 'm-3'
        assert_almost_equal(self.ug.data[0, 0], [1e6])

    def test_change_units_throws_ValueError_for_invalid_units_or_conversion(self):
        # Invalid units on the ungridded data
        with assert_raises(ValueError):
            self.ug.units = 'cm-3 stp'
            self.ug.convert_units('m-3')

        # Invalid units to convert to
        with assert_raises(ValueError):
            self.ug.units = 'cm-3'
            self.ug.convert_units('foo')

        # Invalid conversion
        with assert_raises(ValueError):
            self.ug.units = 'cm-3'
            self.ug.convert_units('kg')



from nose.tools import eq_, istest, raises

from cis.data_io.hdf import _read_hdf4
from cis.exceptions import InvalidVariableError
from cis.test.integration_test_data import *


@istest
@skip_pyhdf
@raises(IOError)
def should_raise_io_error_with_non_hdf_file():
    _read_hdf4(valid_cloud_cci_filename, valid_cloud_cci_variable)


@istest
@skip_pyhdf
def test_read_hdf4():
    filename = escape_colons(valid_hdf_sd_file)
    sds, vds = _read_hdf4(filename, ['Solution_Ocean', 'Path_Radiance_Land', 'Mean_Reflectance_Land'])

    # VD variable are listed in the VD part of the tuple, but not in the SD part
    eq_(True, 'Solution_Ocean' in vds)
    eq_(False, 'Solution_Ocean' in sds)

    # SD variable are listed in the SD part of the tuple, but not in the VD part
    eq_(True, 'Path_Radiance_Land' in sds)
    eq_(False, 'Path_Radiance_Land' in vds)
    eq_(True, 'Mean_Reflectance_Land' in sds)
    eq_(False, 'Mean_Reflectance_Land' in vds)


@istest
@skip_pyhdf
@raises(InvalidVariableError)
def test_that_cannot_read_unknown_variables():
    filename = escape_colons(valid_hdf_sd_file)
    sds, vds = _read_hdf4(filename, ['athing', 'unechose', 'einding'])


@istest
@skip_pyhdf
@raises(InvalidVariableError)
def test_that_cannot_read_unknown_variables_and_valid_variables():
    filename = escape_colons(valid_hdf_sd_file)
    sds, vds = _read_hdf4(filename, ['someBizarreVariableNobodyKnowsAbout', 'Solution_Ocean', 'Path_Radiance_Land',
                                      'Mean_Reflectance_Land'])



"""
Basic scatter plots, against either one or two coordinates.
"""
from cis.plotting.genericplot import GenericPlot, Generic2DPlot


class ScatterPlot(GenericPlot):

    def __call__(self, ax):
        """
        Plots one set of scatter points
        """
        # Translate our argument names to mpl line kwargs
        if self.itemwidth is not None:
            self.mplkwargs["markersize"] = self.itemwidth

        if self.itemstyle is not None:
            self.mplkwargs["marker"] = self.itemstyle

        if self.edgecolor is not None:
            self.mplkwargs["edgecolors"] = self.edgecolor

        if self.color is not None:
            self.mplkwargs["c"] = self.color

        # By plotting the data as flat arrays with a line style of 'o' we get a scatter plot - with the added
        #  benefit of not having to keep track of the layers ourselves.
        ax.plot(self.x.flatten(), self.data.flatten(), 'o', *self.mplargs, **self.mplkwargs)

        super(ScatterPlot, self).__call__(ax)


class ScatterPlot2D(Generic2DPlot):

    def __call__(self, ax):
        """
        Plots one set of scatter points with the colour determined by the data
        """
        if self.itemwidth is not None:
            self.mplkwargs["s"] = self.itemwidth

        if self.itemstyle is not None:
            self.mplkwargs["marker"] = self.itemstyle

        if self.edgecolor is not None:
            self.mplkwargs["edgecolors"] = self.edgecolor
        else:
            # For 2D scatter plots set the edgecolors off by default
            self.mplkwargs["edgecolors"] = None

        self.mplkwargs["c"] = self.data

        self.mappable = ax.scatter(self.x, self.y, *self.mplargs, **self.mplkwargs)

        super(ScatterPlot2D, self).__call__(ax)



import logging

defaultdeletechars = """~!@#$%^&*=+~\|]}[{'; /?.>,<"""

AERONET_HEADER_LENGTH = {"AERONET-SDA/2": 5, "AERONET/2": 5, "MAN-SDA/2": 5, "MAN/2": 5,
                         "AERONET-SDA/3": 7, "AERONET/3": 7}
AERONET_MISSING_VALUE = {"AERONET-SDA/2": 'N/A', "AERONET/2": 'N/A', "MAN-SDA/2": -999.0, "MAN/2": (-999.0, -10000),
                         "AERONET-SDA/3": -999.0, "AERONET/3": -999.0}

V2_HEADER = "Version 2 Direct Sun Algorithm"
V3_HEADER = "AERONET Version 3"

AERONET_COORDINATE_RENAME = {
    "Date(dd:mm:yy)" : "date",
    "Date(dd-mm-yy)" : "date",
    "Date(dd:mm:yyyy)" : "date",
    "Date(dd-mm-yyyy)" : "date",
    "Date_(dd:mm:yy)" : "date",
    "Date_(dd-mm-yy)" : "date",
    "Date_(dd:mm:yyyy)" : "date",
    "Date_(dd-mm-yyyy)" : "date",
    "Time(hh:mm:ss)" : "time",
    "Time(hh-mm-ss)" : "time",
    "Time_(hh:mm:ss)" : "time",
    "Time_(hh-mm-ss)" : "time",
    "Latitude" : "latitude",
    "Longitude" : "longitude",
    "Latitude(Degrees)": "latitude",
    "Longitude(Degrees)": "longitude",
    "Site_Latitude(Degrees)" : "latitude",
    "Site_Longitude(Degrees)" : "longitude",
    "Site_Elevation(m)" : "altitude",
    "Elevation(m)": "altitude"
}


def get_slice_of_lines_from_file(filename, start, end):
    """Grab a subset of lines from a file, defined using slice-style start:end.

    This uses less memory than the equivalent linecache.getline()."""
    from itertools import islice
    with open(filename) as fileobj:
        return list(islice(fileobj, start, end))


def get_aeronet_version(filename):
    """
    Classifies the format of an Aeronet file based on it's header.
    :param filename: Full path to the file to read.
    :return str: AERONET or MAN, followed by an optional -SDA, and /# where # is the version number.
    """
    from cis.exceptions import FileFormatError

    first_line, second_line = get_slice_of_lines_from_file(filename, 0, 2)

    man = "Maritime Aerosol Network" in first_line
    sda = "SDA" in first_line

    if man:
        return "MAN-SDA/2" if sda else "MAN/2"

    if first_line.startswith(V3_HEADER):
        return "AERONET-SDA/3" if sda else "AERONET/3"

    if second_line.startswith(V2_HEADER):
        return "AERONET-SDA/2" if sda else "AERONET/2"

    raise FileFormatError(["Unable to determine Aeronet file version", filename],
                          "Unable to determine Aeronet file version " + filename)


def get_aeronet_file_variables(filename, version=None):
    """
    Return a list of valid Aeronet file variables with invalid characters removed. We need to remove invalid characters
    primarily for writing back out to CF-compliant NetCDF.
    :param filename: Full path to the file to read
    :return: A list of Aeronet variable names in the order they appear in the file
    """
    from collections import Counter

    if version is None:
        version = get_aeronet_version(filename)

    try:
        first_line, second_line = get_slice_of_lines_from_file(filename, AERONET_HEADER_LENGTH[version]-1,
                                                               AERONET_HEADER_LENGTH[version]+1)
    except ValueError:
        # Aeronet files can, for some reason, contain no data
        return []

    variables = first_line.replace("\n", "").split(",")

    # The SDA files don't list all of the columns
    if variables[-1] == "Exact_Wavelengths_for_Input_AOD(um)" or variables[-1] == "Exact_Wavelengths_for_Input_AOD(nm)":
        original_name = variables[-1]

        # Find the number of wavelengths from the first data line
        values = second_line.split(",")
        n_wavelengths = int(values[len(variables)-2])
        for i in range(n_wavelengths-1):
            variables.append(original_name)

    repeated_items = {var:-1 for var, num in Counter(variables).items() if num > 1}

    final_variables = []
    for var in variables:
        # Add a numerical counter to repeated variable names
        if var in repeated_items:
            repeated_items[var] += 1
            var += ".{}".format(repeated_items[var])

        # Remove nonstandard characters
        for char in defaultdeletechars:
            var = var.replace(char, "")

        var = var.strip()

        # Enforce standardised names for the coordinate fields
        try:
            final_variables.append(AERONET_COORDINATE_RENAME[var])
        except KeyError:
            final_variables.append(var)

    return final_variables


def load_multiple_aeronet(filenames, variables=None):
    from cis.utils import add_element_to_list_in_dict, concatenate

    adata = {}

    for filename in filenames:
        logging.debug("reading file: " + filename)

        # reading in all variables into a dictionary:
        # a_dict, key: variable name, value: list of masked arrays
        a_dict = load_aeronet(filename, variables)
        for var in list(a_dict.keys()):
            add_element_to_list_in_dict(adata, var, a_dict[var])

    for var in list(adata.keys()):
        adata[var] = concatenate(adata[var])

    return adata


def load_aeronet(filename, variables=None):
    """
    Loads aeronet csv file.

    :param filename: data file name
    :param variables: A list of variables to return
    :return: A dictionary of variables names and numpy arrays containing the data for that variable
    """
    from cis.exceptions import InvalidVariableError
    from cis.time_util import cis_standard_time_unit
    from numpy.ma import masked_invalid
    from pandas import read_csv, to_datetime

    version = get_aeronet_version(filename)
    ordered_vars = get_aeronet_file_variables(filename, version)
    if len(ordered_vars) == 0:
        return {}

    # Load all available geolocation information and any requested variables
    cols = [var for var in ("date", "time", "latitude", "longitude", "altitude") if var in ordered_vars]
    if cols is not None and variables is not None:
        cols.extend(variables)

    dtypes = {var:'str' if var in ("date", "time") else "float" for var in cols}

    try:
        rawd = read_csv(filename, sep=",", header=AERONET_HEADER_LENGTH[version]-1, names=ordered_vars,
                        index_col=False, usecols=cols, na_values=AERONET_MISSING_VALUE[version], dtype=dtypes,
                        parse_dates={"datetime":["date", "time"]}, infer_datetime_format=True, dayfirst=True,
                        error_bad_lines=False, warn_bad_lines=True, #low_memory="All_Sites_Times_All_Points" in filename
        )
    except ValueError:
        raise InvalidVariableError("{} not available in {}".format(variables, filename))

    # Empty file
    if rawd.shape[0] == 0:
        return {"datetime":[], "latitude":[], "longitude":[], "altitude":[]}

    # Convert pandas Timestamps into CIS standard numbers
    rawd["datetime"] = [cis_standard_time_unit.date2num(timestamp.to_pydatetime())
                        for timestamp in to_datetime(rawd["datetime"], format='%d:%m:%Y %H:%M:%S')]

    # Add position metadata that isn't listed in every line for some formats
    if version.startswith("MAN"):
        rawd["altitude"] = 0.

    elif version.endswith("2"):
        metadata = get_file_metadata(filename)
        rawd["longitude"] = float(metadata.misc[2][1].split("=")[1])
        rawd["latitude"] = float(metadata.misc[2][2].split("=")[1])
        rawd["altitude"] = float(metadata.misc[2][3].split("=")[1])

    return {var : masked_invalid(arr) for var, arr in rawd.items()}


def get_file_metadata(filename, variable='', shape=None):
    file = open(filename)
    from cis.data_io.ungridded_data import Metadata

    if variable is None:
        variable = ''
    version = get_aeronet_version(filename)
    metadata = Metadata(name=variable, long_name=variable, shape=shape)
    lines = []
    for i in range(0, AERONET_HEADER_LENGTH[version]-1):
        lines.append(file.readline().replace("\n", "").split(","))
    metadata.misc = lines
    return metadata



import numpy as np

from cis.collocation.kdtree import HaversineDistanceKDTree
from cis.data_io.hyperpoint import HyperPoint


def create_index(data, leafsize=10):
    """
    Creates the k-D tree index.

    :param data: list of HyperPoints to index
    """
    spatial_points = data[['latitude', 'longitude']]
    if hasattr(data, 'data'):
        mask = np.ma.getmask(data.data).ravel()
    else:
        mask = None
    return HaversineDistanceKDTree(spatial_points, mask=mask, leafsize=leafsize)


class HaversineDistanceKDTreeIndex(object):
    """k-D tree index that can be used to query using distance along the Earth's surface.
    """
    def __init__(self):
        self.index = None

    def index_data(self, points, data, coord_map, leafsize=10):
        """
        Creates the k-D tree index.

        :param points: (not used) sample points
        :param data: list of HyperPoints to index
        :param coord_map: (not used) list of tuples relating index in HyperPoint
                          to index in sample point coords and in coords to be output
        """
        try:
            self.index = create_index(data, leafsize=leafsize)
        except KeyError:
            pass # Unable to create index

    def find_nearest_point(self, point):
        """Finds the indexed point nearest to a specified point.
        :param point: point for which the nearest point is required
        :return: index in data of closest point
        """
        query_pt = point[['latitude', 'longitude']]
        (distances, indices) = self.index.query(query_pt)
        return indices

    def find_points_within_distance(self, point, distance):
        """Finds the points within a specified distance of a specified point.
        :param point: reference point
        :param distance: distance in kilometres
        :return: list indices in data of points
        """
        query_pt = [[point.latitude, point.longitude]]
        return self.index.query_ball_point(query_pt, distance)[0]

    def find_points_within_distance_sample(self, sample, distance):
        """Finds the points within a specified distance of a specified point.
        :param sample: the sample points
        :param distance: distance in kilometres
        :return list of lists:
        For each element ``self.data[i]`` of this tree, ``results[i]`` is a
            list of the indices of its neighbors in ``other.data``.
        """
        return create_index(sample).query_ball_tree(self.index, distance)



"""
 Module to test the collocation routines
"""
from cis.data_io.hyperpoint import HyperPoint
from nose.tools import assert_almost_equal, istest, eq_


@istest
def test_same_point_in_space_and_time_with_points_with_different_values():
    assert (HyperPoint(lat=10.0, lon=50.0, val=14.4).same_point_in_space_and_time(
        HyperPoint(lat=10.0, lon=50.0, val=15.1)))


@istest
def test_is_not_same_point_in_space_and_time_with_points_with_different_values():
    assert (not HyperPoint(lat=10.0, lon=51.0, val=14.4).same_point_in_space_and_time(
        HyperPoint(lat=10.0, lon=50.0, val=15.1)))


@istest
def test_same_point_in_space_with_points_with_different_values():
    assert (HyperPoint(lat=10.0, lon=50.0, alt=10.0, val=14.4).same_point_in_space(
        HyperPoint(lat=10.0, lon=50.0, alt=10.0, val=15.1)))


@istest
def test_is_notsame_point_in_space_with_points_with_different_values():
    assert (not HyperPoint(lat=10.0, lon=50.0, alt=10.0, val=14.4).same_point_in_space(
        HyperPoint(lat=10.0, lon=50.0, alt=11.0, val=15.1)))


@istest
def test_same_point_in_space_with_pressure_not_altitude_with_points_with_different_values():
    assert (HyperPoint(lat=10.0, lon=50.0, pres=10.0, val=14.4).same_point_in_space(
        HyperPoint(lat=10.0, lon=50.0, pres=10.0, val=15.1)))


@istest
def test_is_not_same_point_in_space_with_pressure_not_altitude_with_points_with_different_values():
    assert (not HyperPoint(lat=10.0, lon=50.0, pres=10.0, val=14.4).same_point_in_space(
        HyperPoint(lat=10.0, lon=50.0, pres=11.0, val=15.1)))


@istest
def test_same_point_in_space_with_pressure_and_altitude_with_points_with_different_values():
    assert (HyperPoint(lat=10.0, lon=50.0, alt=5.0, pres=10.0, val=14.4).same_point_in_space(
        HyperPoint(lat=10.0, lon=50.0, alt=5.0, pres=10.0, val=15.1)))


@istest
def test_is_not_same_point_in_space_with_pressure_and_altitude_with_points_with_different_values():
    assert (not HyperPoint(lat=10.0, lon=50.0, alt=5.0, pres=10.0, val=14.4).same_point_in_space(
        HyperPoint(lat=10.0, lon=50.0, alt=6.0, pres=11.0, val=15.1)))


@istest
def test_same_point_in_space_with_points_with_different_times():
    assert (HyperPoint(lat=10.0, lon=50.0, alt=10.0, t=500.0).same_point_in_space(
        HyperPoint(lat=10.0, lon=50.0, alt=10.0, t=501.0)))


@istest
def test_is_not_same_point_in_space_with_points_with_different_times():
    assert (not HyperPoint(lat=10.0, lon=50.0, alt=10.0, t=500.0).same_point_in_space(
        HyperPoint(lat=11.0, lon=51.0, alt=10.0, t=501.0)))


@istest
def test_same_point_in_space_with_pressure_not_altitude_with_points_with_different_times():
    assert (HyperPoint(lat=10.0, lon=50.0, pres=10.0, t=500.0).same_point_in_space(
        HyperPoint(lat=10.0, lon=50.0, pres=10.0, t=501.0)))


@istest
def test_is_not_same_point_in_space_with_pressure_not_altitude_with_points_with_different_times():
    assert (not HyperPoint(lat=10.0, lon=50.0, pres=10.0, t=500.0).same_point_in_space(
        HyperPoint(lat=10.0, lon=50.0, pres=11.0, t=501.0)))


@istest
def test_same_point_in_space_with_pressure_and_altitude_with_points_with_different_times():
    assert (HyperPoint(lat=10.0, lon=50.0, alt=5.0, pres=10.0, t=500.0).same_point_in_space(
        HyperPoint(lat=10.0, lon=50.0, alt=5.0, pres=10.0, t=501.0)))


@istest
def test_is_not_same_point_in_space_with_pressure_and_altitude_with_points_with_different_times():
    assert (not HyperPoint(lat=10.0, lon=50.0, alt=5.0, pres=10.0, t=500.0).same_point_in_space(
        HyperPoint(lat=10.0, lon=50.0, alt=5.0, pres=11.0, t=501.0)))


@istest
def test_same_point_in_time_with_points_with_different_values():
    assert (HyperPoint(lat=10.0, lon=50.0, alt=5.0, t=500.0, val=14.4).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, alt=5.0, t=500.0, val=15.1)))


@istest
def test_is_not_same_point_in_time_with_points_with_different_values():
    assert (not HyperPoint(lat=10.0, lon=50.0, alt=5.0, t=500.0, val=14.4).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, alt=5.0, t=501.0, val=15.1)))


@istest
def test_same_point_in_time_with_pressure_not_altitude_with_points_with_different_values():
    assert (HyperPoint(lat=10.0, lon=50.0, pres=5.0, t=500.0, val=14.4).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, pres=5.0, t=500.0, val=15.1)))


@istest
def test_is_not_same_point_in_time_with_pressure_not_altitude_with_points_with_different_values():
    assert (not HyperPoint(lat=10.0, lon=50.0, pres=5.0, t=500.0, val=14.4).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, pres=5.0, t=501.0, val=15.1)))


@istest
def test_same_point_in_time_with_pressure_and_altitude_with_points_with_different_values():
    assert (HyperPoint(lat=10.0, lon=50.0, alt=10.0, pres=5.0, t=500.0, val=14.4).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, alt=10.0, pres=5.0, t=500.0, val=15.1)))


@istest
def test_is_not_same_point_in_time_with_pressure_and_altitude_with_points_with_different_values():
    assert (not HyperPoint(lat=10.0, lon=50.0, alt=10.0, pres=5.0, t=500.0, val=14.4).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, alt=10.0, pres=5.0, t=501.0, val=15.1)))


@istest
def test_same_point_in_time_with_points_with_different_spatial_coords():
    assert (HyperPoint(lat=10.0, lon=50.0, alt=5.0, t=500.0).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, alt=10.0, t=500.0)))


@istest
def test_is_not_same_point_in_time_with_points_with_different_spatial_coords():
    assert (not HyperPoint(lat=10.0, lon=50.0, alt=5.0, t=500.0).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, alt=10.0, t=501.0)))


@istest
def test_same_point_in_time_with_pressure_not_altitude_with_points_with_different_spatial_coords():
    assert (HyperPoint(lat=10.0, lon=50.0, pres=5.0, t=500.0).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, pres=10.0, t=500.0)))


@istest
def test_is_not_same_point_in_time_with_pressure_not_altitude_with_points_with_different_spatial_coords():
    assert (not HyperPoint(lat=10.0, lon=50.0, pres=5.0, t=500.0).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, pres=10.0, t=501.0)))


@istest
def test_same_point_in_time_with_pressure_and_altitude_with_points_with_different_spatial_coords():
    assert (HyperPoint(lat=10.0, lon=50.0, alt=5.0, pres=4.0, t=500.0).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, alt=10.0, pres=9.0, t=500.0)))


@istest
def test_is_not_same_point_in_time_with_pressure_and_altitude_with_points_with_different_spatial_coords():
    assert (not HyperPoint(lat=10.0, lon=50.0, alt=5.0, pres=4.0, t=500.0).same_point_in_time(
        HyperPoint(lat=10.0, lon=50.0, alt=10.0, pres=9.0, t=501.0)))


@istest
def can_get_valid_coord_tuple_lat():
    eq_(HyperPoint(10).get_coord_tuple(), [('latitude', 10)])


@istest
def can_get_valid_coord_tuple_lat_lon():
    eq_(HyperPoint(10, 12).get_coord_tuple(), [('latitude', 10), ('longitude', 12)])


@istest
def check_furthest_point_on_equator():
    eq_(HyperPoint(0, 0).furthest_point_from(), HyperPoint(0, 180))


@istest
def check_furthest_point_on_pole():
    eq_(HyperPoint(90, 0).furthest_point_from(), HyperPoint(-90, 180))


@istest
def check_furthest_point():
    eq_(HyperPoint(51, 0).furthest_point_from(), HyperPoint(-51, 180))


@istest
def check_dist_between_2d_points_on_equator():
    assert_almost_equal(HyperPoint(0, 0).haversine_dist(HyperPoint(0, 1)), 111.317, places=3)


@istest
def check_dist_between_2d_points_at_pole():
    assert_almost_equal(HyperPoint(90, 0).haversine_dist(HyperPoint(90, 1)), 0.0000, places=5)


@istest
def check_dist_between_opposite_2d_points():
    """
        The distance between two points on opposite sides of the globe should be half the circumfrence of the globe
    """
    import math
    R_E = 6378  # Radius of the earth in km
    max_dist = math.pi * R_E
    assert_almost_equal(HyperPoint(0, 0).haversine_dist(HyperPoint(0, 180)), max_dist, places=3)
    assert_almost_equal(HyperPoint(90, 0).haversine_dist(HyperPoint(-90, 0)), max_dist, places=3)
    assert_almost_equal(HyperPoint(51, 0).haversine_dist(HyperPoint(-51, 180)), max_dist, places=3)


@istest
def check_compdist_returns_true_when_p2_is_closer_than_p1():
    assert (HyperPoint(0, 0).compdist(HyperPoint(4, 4), HyperPoint(3, 3)))


@istest
def check_compdist_returns_false_when_p1_is_closer_than_p2():
    assert (not HyperPoint(0, 0).compdist(HyperPoint(3, 3), HyperPoint(4, 4)))


@istest
def check_compdist_returns_false_when_p1_is_the_same_as_p2():
    assert (not HyperPoint(0, 0).compdist(HyperPoint(3, 3), HyperPoint(3, 3)))
    assert (not HyperPoint(0, 0).compdist(HyperPoint(-3, 3), HyperPoint(3, 3)))
    assert (not HyperPoint(0, 0).compdist(HyperPoint(3, -3), HyperPoint(3, 3)))
    assert (not HyperPoint(0, 0).compdist(HyperPoint(-3, -3), HyperPoint(3, 3)))
    assert (not HyperPoint(-2.5, 0).compdist(HyperPoint(-5, 0), HyperPoint(0, 0)))
    assert (not HyperPoint(0, -2.5).compdist(HyperPoint(0, -5), HyperPoint(0, 0)))
    assert (not HyperPoint(-2.5, -2.5).compdist(HyperPoint(-5, -5), HyperPoint(0, 0)))



from abc import ABCMeta, abstractmethod
import six
from cis.utils import index_iterator_for_non_masked_data, index_iterator_nditer
import numpy as np


@six.add_metaclass(ABCMeta)
class Collocator(object):
    """
    Class which provides a method for performing collocation. This just defines the interface which
    the subclasses must implement.
    """
    def __init__(self, fill_value=None, var_name='', var_long_name='', var_units='',
                 missing_data_for_missing_sample=False):
        """
        Initialise the fill_value, missing data flag and variable attributes.

        :param fill_value: The value to use when the kernel is unable to return a value. The default is NaN.
        :param var_name: The name of the variable to use when creating the output data object
        :param var_long_name: The long name of the variable to use when creating the output data object
        :param var_units: The units of the variable to use when creating the output data object
        :param missing_data_for_missing_sample: If True then sample points which have a missing value will result in
         data points with a missing value - regardless of the collocation result. The default is False.
        :return:
        """
        self.fill_value = float(fill_value) if fill_value is not None else np.nan
        self.var_name = var_name
        self.var_long_name = var_long_name
        self.var_units = var_units
        self.missing_data_for_missing_sample = missing_data_for_missing_sample

    @abstractmethod
    def collocate(self, points, data, constraint, kernel):
        """
        The method is responsible for setting up and running the collocation. It should take a set of data and map that
        onto the given (sample) points using the constraint and kernel provided.

        :param points: A set of sample points onto which we will collocate some other 'data'
        :param data: Some other data to be collocated onto the 'points'
        :param constraint: A :class:`.Constraint` instance which provides a :meth:`.Constraint.constrain_points` method,
         and optionally an :meth:`.Constraint.get_iterator` method
        :param kernel: A :class:`.Kernel` instance which provides a :meth:`.Kernel.get_value` method
        :return: One or more :class:`.CommonData` (or subclasses of) objects whose coordinates lie on the points
         defined above.
        """


class Kernel(object):
    """
    Class which provides a method for taking a number of points and returning one value. For example a nearest
    neighbour algorithm or sort algorithm or mean. This just defines the interface which the subclasses must implement.
    """
    __metaclass__ = ABCMeta

    #: The number of values the :meth:`.Kernel.get_value` should be expected to return
    #: (i.e. the length of the return list).
    return_size = 1

    @abstractmethod
    def get_value(self, point, data):
        """
        This method should return a single value (if :attr:`.Kernel.return_size` is 1) or a list of n values (if
        :attr:`.Kernel.return_size` is n) based on some calculation on the data given a single point.

        The data is deliberately left unspecified in the interface as it may be any type of data, however it is expected
        that each implementation will only work with a specific type of data (gridded, ungridded etc.) Note that this
        method will be called for every sample point and so could become a bottleneck for calculations, it is advisable
        to make it as quick as is practical. If this method is unable to provide a value (for example if no data points
        were given) a ValueError should be thrown.

        :param point: A single HyperPoint
        :param data: A set of data points to reduce to a single value
        :return: For return_size=1 a single value (number) otherwise a list of return values, which represents some
            operation on the points provided
        :raises ValueError: When the method is unable to return a value
        """

    def get_variable_details(self, var_name, var_long_name, var_standard_name, var_units):
        """Returns the details of all variables to be created from the outputs of a kernel.

        :param str var_name: Base variable name
        :param str var_long_name: Base variable long name
        :param str var_standard_name: Base variable standard_name
        :param str var_units: Base variable units
        :return: Tuple of tuples, each containing (variable name, variable long name, variable units)
        """
        return ((var_name, var_long_name, var_standard_name, var_units), )


class AbstractDataOnlyKernel(Kernel):
    """
    A Kernel that can work on data only, e.g. mean only requires the data values to calculate the mean, not the sampling
    point.
    """

    __metaclass__ = ABCMeta

    def get_value(self, point, data):
        """
        This method is redundant in the AbstractDataOnlyKernel and only serves as an interface to
        :meth:`.AbstractDataOnlyKernel`, removing the unnecessary point and checking for one or more data points.

        :param point: A single HyperPoint
        :param data: A set of data points to reduce to a single value
        :return: For return_size=1 a single value (number) otherwise a list of returns values, which represents some
            operation on the points provided
        """
        values = data.vals
        if len(values) == 0:
            raise ValueError
        return self.get_value_for_data_only(values)


    @abstractmethod
    def get_value_for_data_only(self, values):
        """
        This method should return a single value (if :attr:`.Kernel.return_size` is 1) or a list of n values
        (if :attr:`.Kernel.return_size` is n) based on some calculation on the the values (a numpy array).

        Note that this method will be called for every sample point in which data can be placed and so could become a
        bottleneck for calculations, it is advisable to make it as quick as is practical. If this method is unable to
        provide a value (for example if no data points were given) a ValueError should be thrown. This method will not
        be called if there are no values to be used for calculations.

        :param values: A numpy array of values (can not be none or empty)
        :return: A single data item if return_size is 1 or a list of items containing :attr:`.Kernel.return_size` items
        :raises ValueError: If there are any problems creating a value
        """


class Constraint(object):
    """
    Class which provides a method for constraining a set of points. A single HyperPoint is given as a reference
    but the data points to be reduced ultimately may be of any type. This just defines the interface which the
    subclasses must implement.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def constrain_points(self, point, data):
        """
        This method should return a subset of the data given a single reference point.
        It is expected that the data returned should be of the same type as that given - but this isn't mandatory. It is
        possible that this function will return zero points (no data), the collocation class is responsible for
        providing a fill_value.

        :param HyperPoint point: A single HyperPoint
        :param data: A set of data points to be reduced
        :return: A reduced set of data points
        """

    def get_iterator(self, missing_data_for_missing_sample, coord_map, coords, data_points, shape, points, output_data):
        """
        Iterator to iterate through the points needed to be calculated.
        The default iterator, iterates through all the sample points calling :meth:`.Constraint.constrain_points` for
        each one.

        :param missing_data_for_missing_sample: If true anywhere there is missing data on the sample then final point is
         missing; otherwise just use the sample
        :param coord_map: Coordinate map - list of tuples of indexes of hyperpoint coord, data coords and output coords
        :param coords: The coordinates to map the data onto
        :param data_points: The (non-masked) data points
        :param shape: Shape of the final data values
        :param points: The original points object, these are the points to collocate
        :param output_data: Output data set
        :return: Iterator which iterates through (sample indices, hyper point and constrained points) to be placed in
         these points
        """
        from cis.collocation.col_implementations import HyperPoint
        if missing_data_for_missing_sample:
            iterator = index_iterator_for_non_masked_data(shape, points)
        else:
            iterator = index_iterator_nditer(shape, output_data[0])

        for indices in iterator:
            hp_values = [None] * HyperPoint.number_standard_names
            for (hpi, ci, shi) in coord_map:
                hp_values[hpi] = coords[ci].points[indices[shi]]

            hp = HyperPoint(*hp_values)
            constrained_points = self.constrain_points(hp, data_points)
            yield indices, hp, constrained_points


class PointConstraint(Constraint):
    """Superclass of constraints acting on sample points.

    The point argument in constrain_points is a HyperPoint.
    """
    __metaclass__ = ABCMeta
    pass


class CellConstraint(Constraint):
    """Superclass of constraints acting on cells surrounding sample points.

    The point argument in constrain_points is a HyperPoint in which the
    coordinate values are of type iris.coords.Cell.
    """
    __metaclass__ = ABCMeta

    def get_iterator(self, missing_data_for_missing_sample, coord_map, coords, data_points, shape, points, output_data):
        from cis.collocation.col_implementations import HyperPoint
        if missing_data_for_missing_sample:
            iterator = index_iterator_for_non_masked_data(shape, points)
        else:
            iterator = index_iterator_nditer(shape, output_data[0])

        for indices in iterator:
            hp_values = [None] * HyperPoint.number_standard_names
            hp_cell_values = [None] * HyperPoint.number_standard_names
            for (hpi, ci, shi) in coord_map:
                hp_values[hpi] = coords[ci].points[indices[shi]]
                hp_cell_values[hpi] = coords[ci].cell(indices[shi])

            hp = HyperPoint(*hp_values)
            constrained_points = self.constrain_points(HyperPoint(*hp_cell_values), data_points)
            yield indices, hp, constrained_points


# noinspection PyAbstractClass
class IndexedConstraint(Constraint):
    """Superclass of constraints that expect points to be referenced by index.
    """
    __metaclass__ = ABCMeta

    def get_iterator(self, missing_data_for_missing_sample, coord_map, coords, data_points, shape, points, output_data):
        from cis.collocation.col_implementations import HyperPoint
        if missing_data_for_missing_sample:
            iterator = index_iterator_for_non_masked_data(shape, points)
        else:
            iterator = index_iterator_nditer(shape, output_data[0])

        for indices in iterator:
            hp_values = [None] * HyperPoint.number_standard_names
            for (hpi, ci, shi) in coord_map:
                hp_values[hpi] = coords[ci].points[indices[shi]]

            hp = HyperPoint(*hp_values)
            constrained_points = self.constrain_points(indices, data_points)
            yield indices, hp, constrained_points


def __get_class(parent_class, name=None):
    """
    Identify the subclass of parent_class to a given name, if specified.
    If the name is not specified, the routine picks a default.

    Note, only the first filename of the list is use here.

    :param parent_class: A base class
    :param name: name of the class to find
    :return: a subclass of the parent_class
    """
    from cis.plugin import find_plugin_classes
    from cis.exceptions import ClassNotFoundError
    import logging

    all_classes = find_plugin_classes(parent_class, 'cis.collocation.col_implementations')
    for cls in all_classes:

        if name is None:
            # Use default
            pass
        else:
            # Method specified directly
            if name == cls.__name__:
                logging.debug("Selected class " + cls.__name__)
                return cls

    raise ClassNotFoundError("Specified "+parent_class.__name__+" subclass cannot be found")


def get_kernel(method=None):
    """
    Top level routine for finding the correct Kernel object.
    :param str method: The kernel method to find - this should be a string which matches the name of one
    of the subclasses of Kernel
    :return: One of Kernel's subclasses
    """
    from cis.exceptions import ClassNotFoundError
    try:
        kernel_cls = __get_class(Kernel, method)
    except ClassNotFoundError:
        raise ValueError("No kernel found matching name: " + str(method))
    return kernel_cls



class DataWriter(object):
    """
    High level class for writing data to a file
    """

    def write_data(self, data, output_file):
        """
        Write data to a file.

        :param CommonData data: Data to write
        :param str output_file: Output file name
        """
        data.save_data(output_file)



"""
Module to test the parser
The parse raises a SystemExit exception with code 2 if it fails to parse.
Each test therefore ignores SystemExit exceptions with code 2 as they are expected.
"""
import argparse
from unittest import TestCase
import os
from mock import patch
from hamcrest import is_, assert_that, contains_inanyorder
from nose.tools import eq_, raises

from cis.parse import parse_args, expand_file_list
from cis.plotting.plot import plot_types


def escape_colons(string):
    import re
    return re.sub(r'([\:])', r'\\\1', string)


class ParseTestFiles(TestCase):
    def setUp(self):
        """
            Create the temporary files necassary for testing
        :return:
        """
        from tempfile import mkdtemp
        import os

        def _make_file(file):
            """
                Creates an empty file and then closes it (similar to 'touch' on linux)
            :param file:
            :return:
            """
            open(file, 'w').close()

        # Create a temporary directory to put our test files in
        self.data_directory = mkdtemp('cis_test_dir')

        # Name all of our test files (with full paths)
        self.test_directory = os.path.join(self.data_directory, 'test_directory_for_parser')
        os.mkdir(self.test_directory)
        self.escaped_test_directory = escape_colons(self.test_directory)

        self.escaped_test_directory_files = []
        self.test_directory_files = []
        for i in range(1, 4):
            test_directory_file = os.path.join(self.test_directory, 'test_file_for_parser_{}'.format(i))
            _make_file(test_directory_file)
            self.escaped_test_directory_files.append(escape_colons(test_directory_file))
            self.test_directory_files.append(test_directory_file)

        single_valid_file = os.path.join(self.data_directory, "single_data_file")
        _make_file(single_valid_file)
        self.single_valid_file = single_valid_file
        self.escaped_single_valid_file = escape_colons(single_valid_file)

        self.multiple_valid_files = []
        self.escaped_multiple_valid_files = []
        for i in range(1, 3):
            valid_file = os.path.join(self.data_directory, 'data_file_{}'.format(i))
            _make_file(valid_file)
            self.multiple_valid_files.append(valid_file)
            self.escaped_multiple_valid_files.append(escape_colons(valid_file))

        self.data_file_wildcard = os.path.join(self.data_directory, 'data_file*')

    def tearDown(self):
        """
            Remove the test directory and everything below it.
        :return:
        """
        from shutil import rmtree
        rmtree(self.data_directory)


class TestParse(ParseTestFiles):
    """
    Generic parser tests not specific to one particular command
    """

    def test_order_is_preserved_when_specifying_individual_files(self):
        parser = argparse.ArgumentParser()
        files = expand_file_list(self.test_directory_files[0] + "," + self.test_directory_files[1], parser)
        eq_(files, [self.test_directory_files[0], self.test_directory_files[1]])
        files = expand_file_list(self.test_directory_files[1] + "," + self.test_directory_files[0], parser)
        eq_(files, [self.test_directory_files[1], self.test_directory_files[0]])

    def test_directories_are_sorted(self):
        parser = argparse.ArgumentParser()
        files = expand_file_list(self.test_directory, parser)
        eq_(files, self.test_directory_files)

    def test_wildcarded_files_are_sorted(self):
        parser = argparse.ArgumentParser()
        files = expand_file_list(os.path.join(self.test_directory,
                                              'test_file_for_parser_*'), parser)
        eq_(files, self.test_directory_files)

    def test_order_is_preserved_when_specifying_files_even_when_wildcards_and_directories_are_specified_too(self):
        parser = argparse.ArgumentParser()
        files = expand_file_list(','.join([self.single_valid_file, self.data_file_wildcard,
                                           self.test_directory]), parser)
        eq_(files, [self.single_valid_file] + self.multiple_valid_files + self.test_directory_files)

    def test_can_specify_one_valid_filename_and_a_directory(self):
        parser = argparse.ArgumentParser()
        files = expand_file_list(self.single_valid_file + ',' + self.test_directory, parser)
        eq_(files, [self.single_valid_file] + self.test_directory_files)

    def test_can_specify_a_directory(self):
        parser = argparse.ArgumentParser()
        files = expand_file_list(self.test_directory, parser)
        eq_(files, self.test_directory_files)

    def test_can_specify_a_file_with_wildcards(self):
        parser = argparse.ArgumentParser()
        files = expand_file_list(os.path.join(self.test_directory, 'test_file_for_parser_*'), parser)
        eq_(files, self.test_directory_files)
        files = expand_file_list(os.path.join(self.test_directory, '*_1'), parser)
        eq_(files, [self.test_directory_files[0]])
        files = expand_file_list(os.path.join(self.test_directory, 'test_file_for_parser_?'), parser)
        eq_(files, self.test_directory_files)
        files = expand_file_list(os.path.join(self.test_directory, 'test_file_for_parser_[0-9]'), parser)
        eq_(files, self.test_directory_files)

    def test_can_specify_one_valid_filename_and_a_wildcarded_file(self):
        parser = argparse.ArgumentParser()
        files = expand_file_list(
            self.single_valid_file + ',' + os.path.join(self.test_directory, 'test_file_for_parser_[0-9]'),
            parser)
        eq_(files, [self.single_valid_file] + self.test_directory_files)

    def test_duplicate_files_are_not_returned_from_expand_file_list(self):
        parser = argparse.ArgumentParser()
        part_1 = os.path.join(self.test_directory, 'test_file_for_parser_1')
        part_2 = os.path.join(self.test_directory, 'test_file_for_parser_[0-9]')
        files = expand_file_list(','.join([part_1, part_2]), parser)
        eq_(files, self.test_directory_files)

    def test_can_specify_one_valid_filename(self):
        args = ["plot", "var:" + self.escaped_test_directory_files[0]]
        parse_args(args)

    def test_should_raise_error_with_one_invalid_filename(self):
        try:
            args = ["plot", "var:invalidfilename"]
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e

    def test_can_specify_more_than_one_valid_filename(self):
        args = ["plot", "var:" + ','.join(self.escaped_test_directory_files[0:1])]
        parse_args(args)

    def test_should_raise_error_with_a_mixture_of_valid_and_invalid_filenames(self):
        try:
            args = ["plot", "var:" + self.escaped_test_directory_files[0] + ',invalidfilename']
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e

    def test_GIVEN_input_contains_output_WHEN_parse_THEN_raises_error(self):
        dummy_cis_out = 'out.nc'
        args_list = [["subset", "var:" + dummy_cis_out, "x=[-180,180]", "-o", dummy_cis_out[:-3]],
                     ["col", "var1,var2:" + dummy_cis_out, self.escaped_single_valid_file + ':collocator=bin',
                      "-o", dummy_cis_out[:-3]],
                     ["col", "var1,var2:" + self.escaped_single_valid_file, dummy_cis_out + ':collocator=bin',
                      "-o", dummy_cis_out[:-3]],
                     ["aggregate", "var:" + dummy_cis_out, "t", "-o", dummy_cis_out[:-3]]]
        for args in args_list:
            try:
                parse_args(args)
                assert False
            except SystemExit as e:
                if e.code != 2:
                    raise e

    def test_GIVEN_output_missing_file_extension_WHEN_parse_THEN_extension_added(self):
        args = ['eval', 'var1,var2:%s:product=cis' % self.escaped_single_valid_file,
                'var1 + var2 + var3', 'units', '-o', 'output_name']
        parsed = parse_args(args)
        assert_that(parsed.output, is_('output_name.nc'))

    def test_GIVEN_output_has_file_extension_WHEN_parse_THEN_extension_not_added(self):
        args = ['eval', 'var1,var2:%s:product=cis' % self.escaped_single_valid_file,
                'var1 + var2 + var3', 'units', '-o', 'output_name.nc']
        parsed = parse_args(args)
        assert_that(parsed.output, is_('output_name.nc'))

    def test_can_escape_variables_with_colons(self):
        args = ['plot', 'my\:var:{0}:product=cis'.format(self.escaped_single_valid_file)]
        parsed = parse_args(args)
        assert_that('my:var' in parsed.datagroups[0]['variables'])

    def test_GIVEN_plot_output_missing_file_extension_WHEN_parse_THEN_extension_added(self):
        args = ['plot', 'var1:%s:product=cis' % self.escaped_single_valid_file,
                '-o', 'output_name']
        parsed = parse_args(args)
        assert_that(parsed.output, is_('output_name.png'))

    def test_output_file_matches_an_existing_file_with_force_overwrite_option(self):
        from cis.parse import _file_already_exists_and_no_overwrite
        from argparse import Namespace
        from tempfile import NamedTemporaryFile

        with NamedTemporaryFile() as tmpfile:
            existing_file = tmpfile.name
            arguments = Namespace(force_overwrite=True)
            # Test output file is the same as input file
            arguments.output = existing_file
            assert not _file_already_exists_and_no_overwrite(arguments)

    def test_output_file_matches_an_existing_file_with_force_overwrite_env(self):
        from cis.parse import _file_already_exists_and_no_overwrite
        from argparse import Namespace
        from tempfile import NamedTemporaryFile

        with NamedTemporaryFile() as tmpfile:
            existing_file = tmpfile.name
            arguments = Namespace(force_overwrite=False)
            # Test output file is the same as input file
            arguments.output = existing_file
            os.environ["CIS_FORCE_OVERWRITE"] = "True"
            assert not _file_already_exists_and_no_overwrite(arguments)
            os.environ["CIS_FORCE_OVERWRITE"] = ""

    def test_output_file_matches_an_existing_file_with_force_overwrite_env_FALSE(self):
        from cis.parse import _file_already_exists_and_no_overwrite
        from argparse import Namespace
        from tempfile import NamedTemporaryFile

        with NamedTemporaryFile() as tmpfile:
            existing_file = tmpfile.name
            arguments = Namespace(force_overwrite=False)
            # Test output file is the same as input file
            arguments.output = existing_file
            os.environ["CIS_FORCE_OVERWRITE"] = "False"
            with patch("six.moves.input", return_value='n') as in_patch:
                # Check we're not going to overwrite the file
                assert _file_already_exists_and_no_overwrite(arguments)
                # Check we didn't have to ask for input to work that out
                assert in_patch.call_count == 0
            os.environ["CIS_FORCE_OVERWRITE"] = ""

    def test_output_file_matches_an_existing_file_with_no_force_overwrite_will_prompt(self):
        from cis.parse import _file_already_exists_and_no_overwrite
        from argparse import Namespace
        from tempfile import NamedTemporaryFile

        with NamedTemporaryFile() as tmpfile:
            existing_file = tmpfile.name
            arguments = Namespace(force_overwrite=False)
            # Test output file is the same as input file
            arguments.output = existing_file

            # Choose yes to overwrite - so return false
            with patch('six.moves.input', return_value='y'):
                assert not _file_already_exists_and_no_overwrite(arguments)

            # Choose no to not overwrite - return True
            with patch('six.moves.input', return_value='n'):
                assert _file_already_exists_and_no_overwrite(arguments)

            # Choose yes, eventually
            with patch('six.moves.input', side_effect=['blah', 'blah', 'y']):
                assert not _file_already_exists_and_no_overwrite(arguments)

            # Choose the default (no)
            with patch('six.moves.input', return_value=''):
                assert _file_already_exists_and_no_overwrite(arguments)

            # Choose the default, eventually
            with patch('six.moves.input', side_effect=['yo', 'nope', '']):
                assert _file_already_exists_and_no_overwrite(arguments)


class TestParsePlot(ParseTestFiles):
    """
    Tests specific to the plot command
    """

    def test_can_specify_valid_chart_type(self):
        args = ["plot", "var:" + self.escaped_test_directory_files[0], "--type",
                list(plot_types.keys())[0]]
        parse_args(args)

    def test_should_raise_error_with_an_invalid_chart_type(self):
        try:
            args = ["plot", "var:" + self.escaped_test_directory_files[0], "--type", "dfgdfgdfgdfgdfgdf"]
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e

    def test_should_raise_error_with_more_than_one_chart_type(self):
        try:
            args = ["plot", "var:" + self.escaped_test_directory_files[0], "--type",
                    list(plot_types.keys())[0], list(plot_types.keys())[1]]
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e

    def test_can_specify_more_than_one_variable(self):
        args = ["plot", "var:" + self.escaped_test_directory_files[0], "var:" + self.escaped_test_directory_files[0]]
        parse_args(args)

    def test_should_raise_error_when_no_variable_is_specified(self):
        try:
            args = ["plot", self.escaped_test_directory_files[0]]
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e

    def test_should_raise_error_with_invalid_line_width(self):
        try:
            args = ["plot", "var:" + self.escaped_test_directory_files[0], "--itemwidth", "4a0"]
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e

    def test_should_raise_error_with_invalid_line_style(self):
        try:
            args = ["plot", "var:" + self.escaped_test_directory_files[0], "--linestyle", "4a0"]
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e


class TestParseEvaluate(ParseTestFiles):
    """
    Tests specific to the evaluate command
    """

    def test_parse_evaluate_single_datagroup(self):
        args = ['eval', 'var1,var2:%s:product=cis' % self.escaped_single_valid_file,
                'var1 + var2 + var3', 'units', '-o', 'output.nc']
        parsed = parse_args(args)
        assert_that(parsed.command, is_('eval'))
        assert_that(parsed.expr, is_('var1 + var2 + var3'))
        assert_that(parsed.units, is_('units'))
        assert_that(parsed.output, is_('output.nc'))
        assert_that(len(parsed.datagroups), is_(1))
        assert_that(parsed.datagroups[0]['filenames'], is_([self.single_valid_file]))
        assert_that(parsed.datagroups[0]['variables'], is_(['var1', 'var2']))
        assert_that(parsed.datagroups[0]['product'], is_('cis'))

    def test_parse_evaluate_multiple_datagroups(self):
        args = ['eval', 'var1,var2:%s' % self.escaped_multiple_valid_files[0],
                'var3:%s' % self.escaped_multiple_valid_files[1], 'var1^var2 / var3', 'units', '-o', 'output.nc']
        parsed = parse_args(args)
        assert_that(parsed.command, is_('eval'))
        assert_that(parsed.expr, is_('var1^var2 / var3'))
        assert_that(parsed.units, is_('units'))
        assert_that(parsed.output, is_('output.nc'))
        assert_that(len(parsed.datagroups), is_(2))
        assert_that(parsed.datagroups[0]['filenames'], is_([self.multiple_valid_files[0]]))
        assert_that(parsed.datagroups[0]['variables'], is_(['var1', 'var2']))
        assert_that(parsed.datagroups[1]['filenames'], is_([self.multiple_valid_files[1]]))
        assert_that(parsed.datagroups[1]['variables'], is_(['var3']))

    def test_parse_evaluate_output_variable(self):
        args = ['eval', 'var1,var2:%s' % self.escaped_single_valid_file,
                'var3:%s' % self.escaped_multiple_valid_files[0], 'var1^var2 / var3', 'units', '-o',
                'out_var:output.nc']
        parsed = parse_args(args)
        assert_that(parsed.output, is_('output.nc'))
        assert_that(parsed.output_var, is_('out_var'))

    def test_parse_evaluate_no_output_variable(self):
        args = ['eval', 'var1,var2:%s' % self.escaped_single_valid_file,
                'var3:%s' % self.escaped_multiple_valid_files[0], 'var1^var2 / var3', 'units', '-o', 'output.nc']
        parsed = parse_args(args)
        assert_that(parsed.output, is_('output.nc'))
        assert_that(parsed.output_var, is_(None))

    def test_parse_evaluate_no_output(self):
        args = ['eval', 'var1,var2:%s' % self.escaped_single_valid_file,
                'var3:%s' % self.escaped_single_valid_file, 'var1^var2 / var3', 'units']
        parsed = parse_args(args)
        assert_that(parsed.output, is_('out.nc'))
        assert_that(parsed.output_var, is_(None))

    def test_parse_evaluate_invalid_output(self):
        args = ['eval', 'var1,var2:%s' % self.escaped_single_valid_file,
                'var3:%s' % self.escaped_multiple_valid_files[0], 'var1^var2 / var3', 'units', '-o', 'var:var:out']
        try:
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e

    def test_parse_evaluate_valid_aliases(self):
        # Should use the given alias or the variable name if not provided
        args = ['eval', 'var1=alias1,var2:%s' % self.escaped_single_valid_file,
                'var3=alias3:%s' % self.escaped_multiple_valid_files[0], 'var1^var2 / var3', 'units', '-o', 'output.nc']
        parsed = parse_args(args)
        assert_that(parsed.datagroups[0]['variables'], is_(['var1', 'var2']))
        assert_that(parsed.datagroups[0]['aliases'], is_(['alias1', 'var2']))
        assert_that(parsed.datagroups[1]['variables'], is_(['var3']))
        assert_that(parsed.datagroups[1]['aliases'], is_(['alias3']))

    def test_parse_evaluate_duplicate_aliases(self):
        args = ['eval', 'var1=alias1,var2=alias1:%s' % self.escaped_single_valid_file,
                'var1^var2 / var3', 'units', '-o', 'output.nc']
        try:
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e

    def test_parse_evaluate_invalid_aliases(self):
        invalid_var_aliases = ['var1=', '=alias', '=', 'var=a=a']
        for var in invalid_var_aliases:
            args = ['eval', '%s:%s' % (var, self.escaped_single_valid_file),
                    'var1^var2 / var3', 'units', '-o', 'output.nc']
            try:
                parse_args(args)
                assert False
            except SystemExit as e:
                if e.code != 2:
                    raise e

    def test_parse_evaluate_missing_units_single_datagroup(self):
        args = ['eval', 'var1,var2:%s' % self.escaped_single_valid_file,
                'var1 + var2']
        try:
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e

    def test_can_specify_attributes(self):
        args = ['eval', 'var1,var2:%s' % self.escaped_single_valid_file, 'var1^var2 / var3', 'units',
                '--attributes', 'att1=val1,att2=val2']
        parsed = parse_args(args)
        assert_that(parsed.attributes, is_({'att1': 'val1',
                                            'att2': 'val2'}))

    def test_can_specify_attributes_shorthand(self):
        args = ['eval', 'var1,var2:%s' % self.escaped_single_valid_file, 'var1^var2 / var3', 'units',
                '-a', 'att1=val1,att2=val2']
        parsed = parse_args(args)
        assert_that(parsed.attributes, is_({'att1': 'val1',
                                            'att2': 'val2'}))

    def test_invalid_attributes_throws_parser_error(self):
        args = ['eval', 'var1,var2:%s' % self.escaped_single_valid_file, 'var1^var2 / var3', 'units',
                '-a']
        attributes = ['att1val1,att2=val2', '=', '=val', 'key=']
        for attr in attributes:
            full_args = args + [attr]
            try:
                parse_args(full_args)
                assert False
            except SystemExit as e:
                if e.code != 2:
                    raise e


class TestParseStats(ParseTestFiles):
    """
    Tests specific to the stats command
    """

    def test_GIVEN_no_output_WHEN_parse_stats_THEN_output_is_None(self):
        args = ['stats', 'var1,var2:%s' % self.escaped_single_valid_file]
        arguments = parse_args(args)
        assert_that(arguments.output, is_(None))

    def test_GIVEN_one_variable_WHEN_parse_stats_THEN_parser_error(self):
        args = ['stats', 'var1:%s' % self.escaped_single_valid_file]
        try:
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e

    def test_GIVEN_more_than_two_variables_WHEN_parse_stats_THEN_parser_error(self):
        args = ['stats', 'var1,var2:%s' % self.escaped_single_valid_file,
                'var3:%s' % self.escaped_single_valid_file]
        try:
            parse_args(args)
            assert False
        except SystemExit as e:
            if e.code != 2:
                raise e

    def test_GIVEN_two_variables_WHEN_parse_stats_THEN_variables_are_in_datagroups(self):
        args = ['stats', 'var1:%s' % self.escaped_single_valid_file, 'var2:%s' % self.escaped_multiple_valid_files[0]]
        arguments = parse_args(args)
        assert_that(arguments.datagroups[0]['filenames'], is_([self.single_valid_file]))
        assert_that(arguments.datagroups[0]['variables'], is_(['var1']))
        assert_that(arguments.datagroups[1]['filenames'], is_([self.multiple_valid_files[0]]))
        assert_that(arguments.datagroups[1]['variables'], is_(['var2']))

    def test_GIVEN_output_file_WHEN_parse_stats_THEN_output_file_in_arguments(self):
        args = ['stats', 'var1,var2:%s' % self.escaped_single_valid_file, '-o', 'output']
        arguments = parse_args(args)
        assert_that(arguments.output, is_('output.nc'))


class TestParseSubset(ParseTestFiles):
    """
    Tests specific to the subset command
    """

    def test_GIVEN_longitude_limits_not_monotonically_increasing_WHEN_subset_THEN_raises_error(self):
        limits = ['x=[270,90]', 'x=[-30,-60]']
        for lim in limits:
            args = ['subset', 'var1:%s' % self.escaped_single_valid_file, lim]
            try:
                parse_args(args)
                assert False
            except SystemExit as e:
                if e.code != 2:
                    raise

    def test_GIVEN_longitude_limits_wider_than_360_WHEN_subset_THEN_raises_error(self):
        limits = ['x=[-180,360]', 'x=[-1,360]']
        for lim in limits:
            args = ['subset', 'var1:%s' % self.escaped_single_valid_file, lim]
            try:
                parse_args(args)
                assert False
            except SystemExit as e:
                if e.code != 2:
                    raise

    def test_GIVEN_longitude_limits_valid_WHEN_subset_THEN_parsed_OK(self):
        limits = ['x=[-10,10]', 'x=[0,360]', 'x=[-180.0,180.0]']
        for lim in limits:
            args = ['subset', 'var1:%s' % self.escaped_single_valid_file, lim]
            parse_args(args)

    def test_GIVEN_subset_command_WHEN_multiple_variables_in_datagroup_THEN_variables_unpacked(self):
        var1, var2 = 'rain', 'snow'
        limits = 'x=[-10,10],y=[40,60]'
        output = 'subset-out'
        product = 'cis'
        args = ["subset", var1 + "," + var2 + ':' + self.escaped_test_directory_files[0] + ':product=' + product,
                limits, '-o', output]
        main_args = parse_args(args)
        dg = main_args.datagroups
        assert_that(len(dg), is_(1))
        assert_that(dg[0]['filenames'], is_([self.test_directory_files[0]]))
        assert_that(dg[0]['product'], is_('cis'))
        assert_that(dg[0]['variables'], contains_inanyorder('rain', 'snow'))


class TestParseAggregate(ParseTestFiles):
    """
    Tests specific to the aggregate command
    """

    def test_GIVEN_aggregate_command_WHEN_multiple_variables_in_datagroup_THEN_variables_unpacked(self):
        var1, var2 = 'rain', 'snow'
        grid = 'x=[-10,10,2]'
        output = 'aggregate-out'
        product = 'cis'
        args = ["aggregate", var1 + "," + var2 + ':' + self.escaped_test_directory_files[0] + ':product=' + product,
                grid, '-o', output]
        main_args = parse_args(args)
        dg = main_args.datagroups
        assert_that(len(dg), is_(1))
        assert_that(dg[0]['filenames'], is_([self.test_directory_files[0]]))
        assert_that(dg[0]['product'], is_('cis'))
        assert_that(dg[0]['variables'], contains_inanyorder('rain', 'snow'))

    def test_GIVEN_longitude_limits_not_monotonically_increasing_WHEN_aggregate_THEN_raises_error(self):
        limits = ['x=[270,90,10]', 'x=[-30,-60,1]']
        for lim in limits:
            args = ['aggregate', 'var1:%s' % self.escaped_single_valid_file, lim]
            try:
                parse_args(args)
                assert False
            except SystemExit as e:
                if e.code != 2:
                    raise

    def test_GIVEN_longitude_limits_wider_than_360_WHEN_aggregate_THEN_raises_error(self):
        limits = ['x=[-180,360,10]', 'x=[-1,360,5]']
        for lim in limits:
            args = ['aggregate', 'var1:%s' % self.escaped_single_valid_file, lim]
            try:
                parse_args(args)
                assert False
            except SystemExit as e:
                if e.code != 2:
                    raise

    def test_GIVEN_longitude_limits_valid_WHEN_aggregate_THEN_parsed_OK(self):
        limits = ['x=[-10,10,1]', 'x=[0,360,10]', 'x=[-180.0,180.0,5]']
        for lim in limits:
            args = ['aggregate', 'var1:%s' % self.escaped_single_valid_file, lim]
            parse_args(args)

    def test_GIVEN_mixed_limits_valid_WHEN_aggregate_THEN_parsed_OK(self):
        limits = ['x=[-180.0,180.0,0.5],y=[-80.0,10.0,0.1]',
                  'x=[-180.0,180.0,0.5],y=[-80.0,10.0,0.1],t=[2008-05-12,2008-05-12,PT15M]']
        for lim in limits:
            args = ['aggregate', 'var1:%s' % self.escaped_single_valid_file, lim]
            parse_args(args)

    def test_output_file_matches_an_input_file(self):
        from cis.parse import _output_file_matches_an_input_file
        from argparse import Namespace
        from tempfile import NamedTemporaryFile

        with NamedTemporaryFile() as tmpfile:
            input_file = tmpfile.name
            arguments = Namespace()
            # Test output file is the same as input file
            arguments.output = input_file
            arguments.datagroups = [{"filenames": [input_file]}]
            assert _output_file_matches_an_input_file(arguments)

            # Test output file is different
            with NamedTemporaryFile() as tmp_out:
                arguments.output = tmp_out.name
                assert not _output_file_matches_an_input_file(arguments)

            # Test output file is different (and doesn't exist)
            arguments.output = 'blah'
            assert not _output_file_matches_an_input_file(arguments)


class TestParseCollocate(ParseTestFiles):
    """
    Tests specific to the collocate command
    """

    def test_GIVEN_collocate_command_WHEN_multiple_variables_in_datagroup_THEN_variables_unpacked(self):
        var1, var2 = 'rain', 'snow'
        output = 'aggregate-out'
        samplegroup = self.escaped_test_directory_files[0] + ':collocator=bin'
        product = 'cis'
        args = ["col", var1 + "," + var2 + ':' + self.escaped_test_directory_files[0] + ':product=' + product,
                samplegroup, '-o', output]
        main_args = parse_args(args)
        dg = main_args.datagroups
        assert_that(len(dg), is_(1))
        assert_that(dg[0]['filenames'], is_([self.test_directory_files[0]]))
        assert_that(dg[0]['product'], is_('cis'))
        assert_that(dg[0]['variables'], contains_inanyorder('rain', 'snow'))

    def test_can_leave_collocator_missing(self):
        var = 'rain'
        samplegroup = self.escaped_test_directory_files[0] + ':variable=rain'
        args = ["col", var + ':' + self.escaped_test_directory_files[0], samplegroup]
        main_args = parse_args(args)
        sg = main_args.samplegroup
        assert_that(not sg.get('collocator', False))
        assert_that(sg['variable'], is_('rain'))

    def test_can_specify_one_valid_samplefile_and_one_complete_datagroup(self):
        args = ["col", "variable:" + self.escaped_test_directory_files[0], self.escaped_test_directory_files[0] +
                ":collocator=col,constraint=con,kernel=nn"]
        args = parse_args(args)
        eq_([self.test_directory_files[0]], args.samplegroup['filenames'])
        eq_(('col', {}), args.samplegroup['collocator'])
        eq_(('con', {}), args.samplegroup['constraint'])
        eq_(('nn', {}), args.samplegroup['kernel'])
        eq_([{'variables': ['variable'], 'filenames': [self.test_directory_files[0]]}], args.datagroups)

    def test_can_specify_one_valid_samplefile_and_one_datafile_without_other_options(self):
        args = ["col", "variable:" + self.escaped_test_directory_files[0], self.escaped_test_directory_files[0] +
                ':collocator=bin']
        args = parse_args(args)
        eq_([self.test_directory_files[0]], args.samplegroup['filenames'])
        eq_(('bin', {}), args.samplegroup['collocator'])
        eq_([{'variables': ['variable'], 'filenames': [self.test_directory_files[0]]}], args.datagroups)

    def test_can_specify_one_valid_samplefile_and_many_datagroups(self):
        args = ["col", "variable1:" + self.escaped_test_directory_files[0],
                "variable2:" + self.escaped_test_directory_files[0],
                "variable3:" + self.escaped_test_directory_files[0],
                self.escaped_test_directory_files[0] + ':variable=variable4,collocator=col,kernel=nn']
        args = parse_args(args)
        eq_([self.test_directory_files[0]], args.samplegroup['filenames'])
        eq_("variable4", args.samplegroup['variable'])
        eq_(('col', {}), args.samplegroup['collocator'])
        eq_(False, args.samplegroup.get('constraint', False))
        eq_(('nn', {}), args.samplegroup['kernel'])
        eq_([self.test_directory_files[0]], args.datagroups[0]['filenames'])
        eq_(["variable1"], args.datagroups[0]['variables'])
        eq_([self.test_directory_files[0]], args.datagroups[1]['filenames'])
        eq_(["variable2"], args.datagroups[1]['variables'])
        eq_([self.test_directory_files[0]], args.datagroups[2]['filenames'])
        eq_(["variable3"], args.datagroups[2]['variables'])

    def test_can_specify_one_valid_samplefile_and_one_datafile_with_internal_options(self):
        args = ["col", "var1:" + self.escaped_test_directory_files[0], self.escaped_test_directory_files[0] +
                ":variable=var2,constraint=SepConstraint[h_sep=1500,v_sep=22000,t_sep=5000],kernel=nn,collocator=bin"]
        args = parse_args(args)
        eq_([self.test_directory_files[0]], args.datagroups[0]['filenames'])
        eq_(["var1"], args.datagroups[0]['variables'])
        eq_([self.test_directory_files[0]], args.samplegroup['filenames'])
        eq_("var2", args.samplegroup['variable'])
        eq_(('SepConstraint', {'h_sep': '1500', 'v_sep': '22000', 't_sep': '5000'}), args.samplegroup['constraint'])
        eq_(('nn', {}), args.samplegroup['kernel'])
        eq_(('bin', {}), args.samplegroup['collocator'])


class TestParseInfo(ParseTestFiles):
    """
    Tests specific to the info command
    """

    def test_GIVEN_info_command_WHEN_single_file_present_THEN_empty_variable_and_product(self):
        args = ["info", self.escaped_test_directory_files[0]]
        main_args = parse_args(args)
        dg = main_args.datagroups
        assert_that(len(dg), is_(1))
        assert_that(dg[0]['filenames'], is_([self.test_directory_files[0]]))
        assert_that(dg[0].get('product', None), is_(None))
        assert_that(dg[0].get('variables', None), is_(None))

    def test_GIVEN_info_command_WHEN_many_files_present_THEN_empty_variable_and_product(self):
        args = ["info", ",".join(self.escaped_test_directory_files)]
        main_args = parse_args(args)
        dg = main_args.datagroups
        assert_that(len(dg), is_(1))
        assert_that(dg[0]['filenames'], is_(self.test_directory_files))
        assert_that(dg[0].get('product', None), is_(None))
        assert_that(dg[0].get('variables', None), is_(None))

    def test_GIVEN_info_command_WHEN_files_and_single_var_present_THEN_single_variable_and_empty_product(self):
        var1 = 'rain'
        args = ["info", var1 + ':' + self.escaped_test_directory_files[0]]
        main_args = parse_args(args)
        dg = main_args.datagroups
        assert_that(len(dg), is_(1))
        assert_that(dg[0]['filenames'], is_([self.test_directory_files[0]]))
        assert_that(dg[0].get('product', None), is_(None))
        assert_that(dg[0].get('variables', None), is_([var1]))

    def test_GIVEN_info_command_WHEN_files_and_many_vars_present_THEN_many_variable_and_empty_product(self):
        var1, var2 = 'rain', 'snow'
        args = ["info", var1 + "," + var2 + ':' + self.escaped_test_directory_files[0]]
        main_args = parse_args(args)
        dg = main_args.datagroups
        assert_that(len(dg), is_(1))
        assert_that(dg[0]['filenames'], is_([self.test_directory_files[0]]))
        assert_that(dg[0].get('product', None), is_(None))
        assert_that(dg[0].get('variables', None), contains_inanyorder('rain', 'snow'))

    def test_GIVEN_info_command_WHEN_files_and_var_and_product_present_THEN_variable_and_product_set(self):
        var1 = 'rain'
        product = 'cis'
        args = ["info", var1 + ':' + self.escaped_test_directory_files[0] + ':product=' + product]
        main_args = parse_args(args)
        dg = main_args.datagroups
        assert_that(len(dg), is_(1))
        assert_that(dg[0]['filenames'], is_([self.test_directory_files[0]]))
        assert_that(dg[0].get('product', None), is_('cis'))
        assert_that(dg[0].get('variables', None), is_([var1]))



#!/bin/env python
"""
Command line interface for the Climate Intercomparison Suite (CIS)
"""
import sys
import traceback
import logging

from cis.data_io.data_reader import DataReader
from cis import __version__, __status__

logger = logging.getLogger(__name__)


def __error_occurred(e):
    """
    Wrapper method used to print error messages and exit the program.

    :param e: An error object or any string
    """
    sys.stderr.write(str(e) + "\n")
    logging.debug(str(e) + "\n")
    logging.debug(traceback.format_exc())
    logging.error(str(e) + " - check cis.log for details\n")
    exit(1)


def plot_cmd(main_arguments):
    """
    Main routine for handling calls to the 'plot' command.
    Reads in the data files specified and passes the rest of the arguments to the plot function.

    :param main_arguments:    The command line arguments
    """
    from cis.plotting.formatted_plot import Plotter
    from cis.data_io.data_reader import DataReader

    data = DataReader().read_datagroups(main_arguments.datagroups)

    # We have to pop off the arguments which plot isn't expecting so that it treats everything else as an mpl kwarg
    main_arguments = vars(main_arguments)
    _ = main_arguments.pop('command')
    _ = main_arguments.pop("quiet")
    _ = main_arguments.pop("verbose")
    _ = main_arguments.pop("force_overwrite")
    _ = main_arguments.pop("output_var", None)

    layer_opts = [{k: v for k, v in d.items() if k not in ['variables', 'filenames', 'product']}
                  for d in main_arguments.pop('datagroups')]
    Plotter(data, layer_opts=layer_opts, **main_arguments)


def info_cmd(main_arguments):
    """
    Main routine for handling calls to the 'info' command.
    Reads in the variables from the data files specified and lists them to stdout if no
    particular variable was specified, otherwise prints detailed information about each
    variable specified

    :param main_arguments:    The command line arguments (minus the info command)
    """
    from cis.info import info
    dg = main_arguments.datagroups[0]
    info(dg['filenames'], dg.get('variables', None), dg.get('product', None), main_arguments.type)


def col_cmd(main_arguments):
    """
    Main routine for handling calls to the collocate ('col') command.

    :param main_arguments:    The command line arguments (minus the col command)
    """
    from cis.collocation.col_framework import get_kernel
    from cis.parse import check_boolean

    # Read the sample data
    missing_data_for_missing_sample = False
    if main_arguments.samplevariable is not None:
        sample_data = DataReader().read_data_list(main_arguments.samplefiles, main_arguments.samplevariable,
                                                  main_arguments.sampleproduct)[0]
        missing_data_for_missing_sample = True
    else:
        sample_data = DataReader().read_coordinates(main_arguments.samplefiles, main_arguments.sampleproduct)

    # Unpack the sample options
    col_name, col_options = main_arguments.samplegroup.get('collocator', ('', {}))
    kern_name, kern_options = main_arguments.samplegroup.get('kernel', ('', {}))

    missing_data_for_missing_sample = check_boolean(col_options.pop('missing_data_for_missing_sample',
                                                                    str(missing_data_for_missing_sample)), logging)

    kernel = get_kernel(kern_name)(**kern_options) if kern_name else None

    for input_group in main_arguments.datagroups:
        # Then collocate each datagroup
        data = DataReader().read_single_datagroup(input_group)
        output = data.collocated_onto(sample_data, how=col_name, kernel=kernel,
                                      missing_data_for_missing_sample=missing_data_for_missing_sample, **col_options)
        output.save_data(main_arguments.output)


def subset_cmd(main_arguments):
    """
    Main routine for handling calls to the subset command.

    :param main_arguments:    The command line arguments (minus the subset command)
    """
    import cis.exceptions as ex

    if len(main_arguments.datagroups) > 1:
        __error_occurred("Subsetting can only be performed on one data group")

    data = DataReader().read_single_datagroup(main_arguments.datagroups[0])

    subset = data.subset(**main_arguments.limits)

    if subset is None:
        # Constraints exclude all data.
        raise ex.NoDataInSubsetError("No output created - constraints exclude all data")

    subset.save_data(main_arguments.output)


def aggregate_cmd(main_arguments):
    """
    Main routine for handling calls to the aggregation command.

    :param main_arguments: The command line arguments (minus the aggregate command)
    """
    import cis.exceptions as ex
    from cis.data_io.gridded_data import GriddedDataList

    if len(main_arguments.datagroups) > 1:
        __error_occurred("Aggregation can only be performed on one data group")
    input_group = main_arguments.datagroups[0]

    data = DataReader().read_single_datagroup(input_group)

    if isinstance(data, GriddedDataList):
        logging.warning("The aggregate command is deprecated for GriddedData and will not be supported in future "
                        "versions of CIS. Please use 'collapse' instead.")
        if any(v is not None for v in main_arguments.grid.values()):
            raise ex.InvalidCommandLineOptionError("Grid specifications are not supported for Gridded aggregation.")
        output = data.collapsed(list(main_arguments.grid.keys()), how=input_group.get("kernel", ''))
    else:
        output = data.aggregate(how=input_group.get("kernel", ''), **main_arguments.grid)

    output.save_data(main_arguments.output)


def collapse_cmd(main_arguments):
    """
    Main routine for handling calls to the collapse command.

    :param main_arguments: The command line arguments (minus the collapse command)
    """
    from cis.data_io.ungridded_data import UngriddedDataList

    if len(main_arguments.datagroups) > 1:
        __error_occurred("Collapse can only be performed on one data group")
    input_group = main_arguments.datagroups[0]

    data = DataReader().read_single_datagroup(input_group)

    if isinstance(data, UngriddedDataList):
        logging.error("The collapse command can only be performed on gridded data. "
                      "Please use 'aggregate' instead.")

    output = data.collapsed(main_arguments.dimensions, how=input_group.get("kernel", ''))

    output.save_data(main_arguments.output)


def evaluate_cmd(main_arguments):
    """
    Main routine for handling calls to the evaluation command

    :param main_arguments: The command line arguments (minus the eval command)
    """
    from cis.evaluate import Calculator
    data_reader = DataReader()
    data_list = data_reader.read_datagroups(main_arguments.datagroups)
    calculator = Calculator()
    result = calculator.evaluate(data_list, main_arguments.expr, main_arguments.output_var,
                                 main_arguments.units, main_arguments.attributes)
    result.save_data(main_arguments.output)


def stats_cmd(main_arguments):
    """
    Main routine for handling calls to the statistics command.

    :param main_arguments: The command line arguments (minus the stats command)
    """
    from cis.stats import StatsAnalyzer
    from cis.data_io.gridded_data import GriddedDataList
    data_reader = DataReader()
    data_list = data_reader.read_datagroups(main_arguments.datagroups)
    analyzer = StatsAnalyzer(*data_list)
    results = analyzer.analyze()
    header = "RESULTS OF STATISTICAL COMPARISON:"
    note = "Compared all points which have non-missing values in both variables"
    header_length = max(len(header), len(note))
    print(header_length * '=')
    print(header)
    print(header_length * '-')
    print(note)
    print(header_length * '=')
    for result in results:
        print(result.pprint())
    if main_arguments.output:
        cubes = GriddedDataList([result.as_cube() for result in results])
        variables = []
        filenames = []
        for datagroup in main_arguments.datagroups:
            variables.extend(datagroup['variables'])
            filenames.extend(datagroup['filenames'])
        history = "Statistical comparison performed using CIS version " + __version__ + \
                  "\n variables: " + str(variables) + \
                  "\n from files: " + str(set(filenames))
        cubes.add_history(history)
        cubes.save_data(main_arguments.output)


def version_cmd(_main_arguments):
    print("Using CIS version: {ver} ({stat})".format(ver=__version__, stat=__status__ ))


commands = {'plot': plot_cmd,
            'info': info_cmd,
            'collocate': col_cmd,
            'aggregate': aggregate_cmd,
            'subset': subset_cmd,
            'collapse': collapse_cmd,
            'eval': evaluate_cmd,
            'stats': stats_cmd,
            'version': version_cmd}


def parse_and_run_arguments(arguments=None):
    """
    Parse and run the arguments
    :param arguments: an arguments list or None to parse all
    """
    from datetime import datetime

    from cis.parse import parse_args

    # parse command line arguments
    arguments = parse_args(arguments)
    command = arguments.command
    logging.debug("CIS started at: " + datetime.now().strftime("%Y-%m-%d %H:%M"))
    logging.debug("Running command: " + command)
    logging.debug("With the following arguments: " + str(arguments))

    # execute command
    cmd = commands[command]
    cmd(arguments)


def main():
    """
    The main method for the program.
    Sets up logging, parses the command line arguments and then calls the appropriate command with its arguments
    """
    import os
    import logging
    from logging import config

    # configure logging
    try:
        logging.config.fileConfig(os.path.join(os.path.dirname(__file__), "logging.conf"))
    except IOError as e:
        # If we don't have permission to write to the log file, all we can do is inform the user
        # All future calls to the logging module will be ignored (?)
        print("WARNING: Unable to write to the log: %s" % e)
    logging.captureWarnings(True)  # to catch warning from 3rd party libraries

    try:
        parse_and_run_arguments()
    except MemoryError:
        __error_occurred("Not enough memory. Please either reduce the amount "
                         "of data, increase the swap space available on your machine or use a machine "
                         "with more memory (for example the JASMIN facility).")
    except Exception as e:
        __error_occurred(e)


if __name__ == '__main__':
    main()



"""
CIS is an open source command-line tool and Python library for easy collocation, visualization, analysis, and
comparison of diverse gridded and ungridded datasets used in the atmospheric sciences.

.. note ::

    The CIS documentation has detailed usage information, including a :doc:`user guide <../index>`
    for new users.

The :func:`read_data` function is a simple way to read a single gridded or ungridded data object (e.g. a NetCDF
variable) from one or more files. CIS will determine the best way to interpret the datafile by comparing the file
signature with the built-in data reading plugins and any user defined plugins. Specifying a particular ``product``
allows the user to override this automatic detection.

The :func:`read_data_list` function is very similar to :func:`read_data` except that it allows the user to specify
more than one variable name. This function returns a list of data objects, either all of which will be gridded, or all
ungridded, but not a mix. For ungridded data lists it is assumed that all objects share the same coordinates.
"""
__author__ = "David Michel, Daniel Wallis, Duncan Watson-Parris, Richard Wilkinson, Ian Bush, Matt Kendall, John Holt"
__version__ = "1.7.3"
__status__ = "Stable"
__website__ = "http://www.cistools.net/"

__all__ = ['read_data', 'read_data_list', 'get_variables']


def read_data(filenames, variable, product=None):
    """
    Read a specific variable from a list of files
    Files can be either gridded or ungridded but not a mix of both.
    First tries to read data as gridded, if that fails, tries as ungridded.

    :param filenames:   The filenames of the files to read. This can be either a single filename as a string, a comma
     separated list, or a :class:`list` of string filenames. Filenames can include directories which will be expanded to
     include all files in that directory, or wildcards such as ``*`` or ``?``.
    :type filenames: string or list
    :param str variable: The variable to read from the files
    :param str product:  The name of the data reading plugin to use to read the data (e.g. ``Cloud_CCI_L2``).
    :return:  The specified data as either a :class:`GriddedData` or :class:`UngriddedData` object.
    """
    data_list = read_data_list(filenames, variable, product)
    if len(data_list) > 1:
        raise ValueError("More than one {} variable found".format(variable))
    return data_list[0]


def read_data_list(filenames, variables, product=None, aliases=None):
    """
    Read multiple data objects from a list of files. Files can be either gridded or ungridded but not a mix of both.

    :param filenames:   The filenames of the files to read. This can be either a single filename as a string, a comma
     separated list, or a :class:`list` of string filenames. Filenames can include directories which will be expanded to
     include all files in that directory, or wildcards such as ``*`` or ``?``.
    :type filenames: string or list
    :param variables: One or more variables to read from the files
    :type variables: string or list
    :param str product: The name of the data reading plugin to use to read the data (e.g. ``Cloud_CCI_L2``).
    :param aliases: List of aliases to put on each variable's data object as an alternative means of identifying them.
    :type aliases: string or list
    :return:  A list of the data read out (either a :class:`GriddedDataList` or :class:`UngriddedDataList` depending on
     the type of data contained in the files)
    """
    from cis.data_io.data_reader import DataReader, expand_filelist
    try:
        file_set = expand_filelist(filenames)
    except ValueError as e:
        raise IOError(e)
    if len(file_set) == 0:
        raise IOError("No files found which match: {}".format(filenames))
    return DataReader().read_data_list(file_set, variables, product, aliases)


def get_variables(filenames, product=None, type=None):
    """
    Get a list of variables names from a list of files. Files can be either gridded or ungridded but not a mix of both.

    :param filenames:   The filenames of the files to read. This can be either a single filename as a string, a comma
     separated list, or a :class:`list` of string filenames. Filenames can include directories which will be expanded to
     include all files in that directory, or wildcards such as ``*`` or ``?``.
    :type filenames: string or list
    :param str product: The name of the data reading plugin to use to read the data (e.g. ``Cloud_CCI_L2``).
    :param str type: The type of HDF data to read, i.e. 'VD' or 'SD'
    :return:  A list of the variables
    """
    from cis.data_io.data_reader import expand_filelist
    from cis.data_io.products.AProduct import get_variables

    try:
        file_set = expand_filelist(filenames)
    except ValueError as e:
        raise IOError(e)
    if len(file_set) == 0:
        raise IOError("No files found which match: {}".format(filenames))
    return get_variables(file_set, product=product, data_type=type)



"""
    Module for the UngriddedData class
"""
import logging
from time import gmtime, strftime
import numpy

import six
from cis.data_io.netcdf import get_data as netcdf_get_data
from cis.data_io.hdf_vd import get_data as hdf_vd_get_data
from cis.data_io.hdf_sd import get_data as hdf_sd_get_data
from cis.data_io.common_data import CommonData, CommonDataList
from cis.data_io.hyperpoint_view import UngriddedHyperPointView
from cis.data_io.write_netcdf import add_data_to_file, write_coordinates
from cis.utils import listify
import cis.maths


class Metadata(object):

    @classmethod
    def from_CubeMetadata(cls, cube_meta):
        return cls(name=cube_meta.var_name, standard_name=cube_meta.standard_name, long_name=cube_meta.long_name,
                   units=str(cube_meta.units), misc=cube_meta.attributes)

    def __init__(self, name='', standard_name='', long_name='', shape=None, units='', range=None, factor=None,
                 offset=None, missing_value=None, history='', misc=None):
        self._name = name

        self._standard_name = ''
        if standard_name:
            self.standard_name = standard_name
        elif name:
            self.standard_name = Metadata.guess_standard_name(name)

        self.long_name = long_name
        self.shape = shape

        self.units = units
        self.range = range
        self.factor = factor
        self.offset = offset
        self.missing_value = missing_value
        self.history = history
        if misc is None:
            self.misc = {}
        else:
            self.misc = misc

    def __eq__(self, other):
        result = NotImplemented

        if isinstance(other, Metadata):
            result = self._name == other._name and \
                     self._standard_name == other._standard_name and \
                     self.long_name == other.long_name and \
                     self.units == other.units
        return result

    # Must supply __ne__, Python does not defer to __eq__ for negative equality
    def __ne__(self, other):
        result = self.__eq__(other)
        if result is not NotImplemented:
            result = not result
        return result

    # Must supply __hash__, Python 3 does not enable it if __eq__ is defined
    def __hash__(self):
        return hash(id(self))

    def summary(self, offset=5):
        """
        Creates a unicode summary of the metadata object

        :param offset: The left hand padding to apply to the text
        :return: The summary
        """
        from datetime import datetime
        string = ''
        string += '{pad:{width}}Long name = {lname}\n'.format(pad=' ', width=offset, lname=self.long_name)
        string += '{pad:{width}}Standard name = {sname}\n'.format(pad=' ', width=offset, sname=self.standard_name)
        string += '{pad:{width}}Units = {units}\n'.format(pad=' ', width=offset, units=self.units)
        string += '{pad:{width}}Missing value = {mval}\n'.format(pad=' ', width=offset, mval=self.missing_value)
        # str(tuple) returns repr(obj) on each item in the tuple, if we have a datetime tuple then we want str(obj)
        #  instead. Just make that ourselves here instead (as a str to avoid the extra quotes if we make a 'real' tuple)
        if isinstance(self.range[0], datetime):
            range_tuple = '({}, {})'.format(*self.range)
        else:
            range_tuple = self.range
        string += '{pad:{width}}Range = {range}\n'.format(pad=' ', width=offset, range=range_tuple)
        string += '{pad:{width}}History = {history}\n'.format(pad=' ', width=offset, history=self.history)
        if self.misc:
            string += '{pad:{width}}Misc attributes: \n'.format(pad=' ', width=offset)
            for k, v in self.misc.items():
                string += '{pad:{width}}{att} = {val}\n'.format(pad=' ', width=offset + 2, att=k.title(), val=v)
        return string

    def __str__(self):
        # six has a decorator for this bit, but it doesn't do errors='replace'.
        if six.PY3:
            return self.summary()
        else:
            return self.summary().encode(errors='replace')

    def __unicode__(self):
        return self.summary()

    @property
    def standard_name(self):
        return self._standard_name

    @standard_name.setter
    def standard_name(self, standard_name):
        from iris.std_names import STD_NAMES
        if standard_name is None or standard_name in STD_NAMES:
            # If the standard name is actually changing from one to another then log the fact
            if self.standard_name is not None \
                    and self.standard_name.strip() != "" \
                    and self.standard_name != standard_name:
                logging.debug("Changing standard name for dataset from '{}' to '{}'".format(self.standard_name,
                                                                                            standard_name))
            self._standard_name = standard_name
        else:
            raise ValueError('%r is not a valid standard_name' % standard_name)

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, units):
        from cf_units import Unit
        if not isinstance(units, Unit):
            try:
                # Try some basic tidying up of unit
                if isinstance(units, six.string_types):
                    if 'since' in units.lower():
                        # Often this time since epoch units are weirdly capitalised, or include extra punctuation
                        units = units.lower().replace("since:", "since").replace(",", "")
                    else:
                        # Replace number with 1 (e.g. #/cm3 == cm-3)
                        units = units.replace('#', '1')
                units = Unit(units)
            except ValueError:
                logging.info("Unable to parse cf-units: {}. Some operations may not be available.".format(units))
        self._units = units

    @staticmethod
    def guess_standard_name(name):
        standard_name = None
        if name.lower().startswith('lat'):
            standard_name = 'latitude'
        elif name.lower().startswith('lon'):
            standard_name = 'longitude'
        elif name.lower().startswith('alt') or name.lower() == 'height':
            standard_name = 'altitude'
        elif name.lower().startswith('pres') or name.lower() == 'air_pressure':
            standard_name = 'air_pressure'
        elif name.lower() == 'time':
            standard_name = 'time'
        return standard_name


# This defines the mappings for each of the ungridded data types to their reading routines, this allows 'lazy loading'
static_mappings = {"SDS": hdf_sd_get_data,
                   "HDF_SDS": hdf_sd_get_data,
                   "VDS": hdf_vd_get_data,
                   "Variable": netcdf_get_data,
                   "_Variable": netcdf_get_data}


class LazyData(object):
    """
        Wrapper (adaptor) class for the different types of possible ungridded data.
    """

    def __init__(self, data, metadata, data_retrieval_callback=None):
        """
        :param data:    The data handler (e.g. SDS instance) for the specific data type, or a numpy array of data
                        This can be a list of data handlers, or a single data handler
        :param metadata: Any associated metadata
        :param data_retrieval_callback: An, optional, method for retrieving data when needed
        """
        from cis.exceptions import InvalidDataTypeError
        from iris.cube import CubeMetadata
        import numpy as np

        self._data_flattened = None

        self.attributes = {}

        self.metadata = Metadata.from_CubeMetadata(metadata) if isinstance(metadata, CubeMetadata) else metadata

        if isinstance(data, np.ndarray):
            # If the data input is a numpy array we can just copy it in and ignore the data_manager
            self._data = data
            self._data_manager = None
            self._post_process()
        else:
            # If the data input wasn't a numpy array we assume it is a data reference (e.g. SDS) and we refer
            #  this as a 'data manager' as it is responsible for getting the actual data.

            self._data = None
            # Although the data can be a list or a single item it's useful to cast it
            #  to a list here to make accessing it consistent
            self._data_manager = listify(data)

            if data_retrieval_callback is not None:
                # Use the given data retrieval method
                self.retrieve_raw_data = data_retrieval_callback
            elif type(self._data_manager[0]).__name__ in static_mappings and \
                    all([type(d).__name__ == type(self._data_manager[0]).__name__ for d in self._data_manager]):
                # Check that we recognise the data manager and that they are all the same

                # Set the retrieve_raw_data method to it's mapped function name
                self.retrieve_raw_data = static_mappings[type(self._data_manager[0]).__name__]
            else:
                raise InvalidDataTypeError

    def name(self, default='unknown'):
        """
        This routine returns the first name property which is not empty out of: standard_name, long_name and var_name.
        If they are all empty it returns the default string (which is 'unknown' by default).
        :return: The name of the data object as a string
        """
        return self.standard_name or self.long_name or self.var_name or default

    @property
    def shape(self):
        return self.metadata.shape

    @shape.setter
    def shape(self, shape):
        self.metadata.shape = shape

    @property
    def long_name(self):
        return self.metadata.long_name

    @long_name.setter
    def long_name(self, long_name):
        self.metadata.long_name = long_name

    @property
    def standard_name(self):
        return self.metadata.standard_name

    @standard_name.setter
    def standard_name(self, standard_name):
        self.metadata.standard_name = standard_name

    @property
    def var_name(self):
        return self.metadata._name

    @var_name.setter
    def var_name(self, var_name):
        self.metadata._name = var_name

    @property
    def units(self):
        return self.metadata.units

    @units.setter
    def units(self, units):
        self.metadata.units = units

    @property
    def data(self):
        """
        This is a getter for the data property. It caches the raw data if it has not already been read.
        Throws a MemoryError when reading for the first time if the data is too large.
        """
        import numpy.ma as ma
        if self._data is None:
            try:
                # If we were given a list of data managers then we need to concatenate them now...
                self._data = self.retrieve_raw_data(self._data_manager[0])
                if len(self._data_manager) > 1:
                    for manager in self._data_manager[1:]:
                        self._data = ma.concatenate((self._data, self.retrieve_raw_data(manager)), axis=0)
                self._post_process()
            except MemoryError:
                raise MemoryError(
                    "Failed to read the ungridded data as there was not enough memory available.\n"
                    "Consider freeing up variables or indexing the cube before getting its data.")
        return self._data

    def _post_process(self):
        """
        Perform a post-processing step on lazy loaded data
        :return: None
        """
        pass

    def __eq__(self, other):
        import numpy as np
        result = NotImplemented

        if isinstance(other, LazyData):
            # Check the metadata
            result = self.metadata == other.metadata

            # Then, if that is OK, check the data
            if result:
                result = np.allclose(self.data, other.data)

        return result

    # Must supply __ne__, Python does not defer to __eq__ for negative equality
    def __ne__(self, other):
        result = self.__eq__(other)
        if result is not NotImplemented:
            result = not result
        return result

    # Must supply __hash__, Python 3 does not enable it if __eq__ is defined
    def __hash__(self):
        return hash(id(self))

    # Maths operator overloads
    __add__ = cis.maths.add
    __radd__ = __add__
    __sub__= cis.maths.subtract
    __mul__ = cis.maths.multiply
    __rmul__ = cis.maths.multiply
    __div__ = cis.maths.divide
    __truediv__ = cis.maths.divide
    __pow__ = cis.maths.exponentiate

    @data.setter
    def data(self, value):
        self._data = value
        self._data_flattened = None

    @property
    def data_flattened(self):
        """Returns a 1D flattened view (or copy, if necessary) of the data.
        """
        if self._data_flattened is None:
            data = self.data
            self._data_flattened = data.ravel()
        return self._data_flattened

    def add_history(self, new_history):
        """Appends to, or creates, the metadata history attribute using the supplied history string.
        The new entry is prefixed with a timestamp.

        :param new_history: history string
        """
        timestamp = strftime("%Y-%m-%dT%H:%M:%SZ ", gmtime())
        if hasattr(self.metadata, 'history') and len(self.metadata.history) > 0:
            self.metadata.history += '\n' + timestamp + new_history
        else:
            self.metadata.history = timestamp + new_history

    def add_attributes(self, attributes):
        """
        Add a variable attribute to this data

        :param attributes: Dictionary of attribute names (keys) and values.
        :return:
        """
        self.attributes.update(attributes)

    def remove_attribute(self, key):
        """
        Remove a variable attribute from this data

        :param key: Attribute key to remove
        :return:
        """
        self.attributes.pop(key, None)

    def save_data(self, output_file):
        logging.info('Saving data to %s' % output_file)
        write_coordinates(self, output_file)
        add_data_to_file(self, output_file)

    def update_shape(self, shape=None):
        if shape:
            self.metadata.shape = shape
        else:
            self.metadata.shape = self.data.shape

    def update_range(self, range=None):
        from cis.time_util import cis_standard_time_unit

        # If the user hasn't specified a range then work it out...
        if not range:
            standard_time = False
            try:
                standard_time = self.units == cis_standard_time_unit
            except ValueError:
                # If UDUNITS can't compare the units then it will raise a ValueError, in which case it's definitely not
                # our standard time
                pass

            try:
                if standard_time:
                    range = (cis_standard_time_unit.num2date(self.data.min()),
                             cis_standard_time_unit.num2date(self.data.max()))
                else:
                    range = (self.data.min(), self.data.max())
            except ValueError as e:
                # If we can't set a range for some reason then just leave it blank
                range = ()

        self.metadata.range = range

    def convert_units(self, new_units):
        """
        Convert units of LazyData object to new_units in place

        :param LazyData ug_data:
        :param cf_units.Unit or str new_units:
        :raises ValueError if new_units can't be converted to standard units, or units are incompatible
        """
        from cf_units import Unit
        if not isinstance(new_units, Unit):
            new_units = Unit(new_units)
        if not isinstance(self.units, Unit):
            # If our units aren't cf_units then they can't be...
            raise ValueError("Unable to convert non-standard LazyData units: {}".format(self.units))
        self.units.convert(self.data, new_units, inplace=True)
        self.units = new_units


class UngriddedData(LazyData, CommonData):
    """
        Wrapper (adaptor) class for the different types of possible ungridded data.
    """

    def __init__(self, data, metadata, coords, data_retrieval_callback=None):
        """
        Constructor

        :param data: The data handler (e.g. SDS instance) for the specific data type, or a numpy array of data.
         This can be a list of data handlers, or a single data handler
        :param metadata: Any associated metadata
        :param coords: A list of the associated Coord objects
        :param data_retrieval_callback: A method for retrieving data when needed
        """
        from cis.data_io.Coord import CoordList, Coord

        if isinstance(coords, list):
            self._coords = CoordList(coords)
        elif isinstance(coords, CoordList):
            self._coords = coords
        elif isinstance(coords, Coord):
            self._coords = CoordList([coords])
        else:
            raise ValueError("Invalid Coords type")

        # TODO Find a cleaner workaround for this, for some reason UDUNITS can not parse 'per kilometer per steradian'
        if str(metadata.units) == 'per kilometer per steradian':
            metadata.units = 'kilometer^-1 steradian^-1'

        super(UngriddedData, self).__init__(data, metadata, data_retrieval_callback)

    @property
    def coords_flattened(self):
        all_coords = self.coords().find_standard_coords()
        return [(c.data_flattened if c is not None else None) for c in all_coords]

    def _post_process(self):
        """
        Perform a post processing step on lazy loaded Ungridded Data.

        :return:
        """
        # Load the data if not already loaded
        if self._data is None:
            data = self.data
        else:
            # Remove any points with missing coordinate values:
            combined_mask = numpy.zeros(self._data.shape, dtype=bool).flatten()
            for coord in self._coords:
                combined_mask |= numpy.ma.getmaskarray(coord.data).flatten()
                if coord.data.dtype != 'object':
                    combined_mask |= numpy.isnan(coord.data).flatten()
                coord.update_shape()
                coord.update_range()
            if combined_mask.any():
                n_points = numpy.count_nonzero(combined_mask)
                logging.warning(
                    "Identified {n_points} point(s) which were missing values for some or all coordinates - "
                    "these points have been removed from the data.".format(n_points=n_points))
                for coord in self._coords:
                    coord.data = numpy.ma.masked_array(coord.data.flatten(), mask=combined_mask).compressed()
                    coord.update_shape()
                    coord.update_range()
                if numpy.ma.is_masked(self._data):
                    new_data_mask = numpy.ma.masked_array(self._data.mask.flatten(), mask=combined_mask).compressed()
                    new_data = numpy.ma.masked_array(self._data.data.flatten(), mask=combined_mask).compressed()
                    self._data = numpy.ma.masked_array(new_data, mask=new_data_mask)
                else:
                    self._data = numpy.ma.masked_array(self._data.flatten(), mask=combined_mask).compressed()
            self.update_shape()
            self.update_range()

    def make_new_with_same_coordinates(self, data=None, var_name=None, standard_name=None,
                                       long_name=None, history=None, units=None, flatten=False):
        """
        Create a new, empty UngriddedData object with the same coordinates as this one.

        :param data: Data to use (if None then defaults to all zeros)
        :param var_name: Variable name
        :param standard_name: Variable CF standard name
        :param long_name: Variable long name
        :param history: Data history string
        :param units: Variable units
        :param flatten: Whether to flatten the data and coordinates (for ungridded data only)
        :return: UngriddedData instance
        """
        if data is None:
            data = numpy.zeros(self.shape)
        metadata = Metadata(name=var_name, standard_name=standard_name,
                            long_name=long_name, history='', units=units)
        if flatten:
            from cis.data_io.Coord import Coord
            data = data.flatten()
            new_coords = []
            for coord in self._coords:
                new_coords.append(Coord(coord.data_flattened, coord.metadata, coord.axis))
        else:
            new_coords = self._coords
        ug_data = UngriddedData(data=data, metadata=metadata, coords=new_coords)
        # Copy the history in separately so it gets the timestamp added.
        if history:
            ug_data.add_history(history)
        return ug_data

    def __getitem__(self, keys):
        """
        Return a COPY of the data with the given slice. We copy to emulate the Iris Cube behaviour
        """
        from copy import deepcopy
        # Create a copy of the slice of each of the coords
        new_coords = []
        for c in self.coords():
            new_coords.append(c[keys])
        # The data is just a new LazyData objects with the sliced data. Note this is a slice of the whole (concatenated)
        #  data, and will lead to post-processing before slicing.
        # TODO: We could be cleverer and figure out the right slice across the various data managers to only read the
        #  right data from disk.
        return UngriddedData(data=self.data[keys].copy(), metadata=deepcopy(self.metadata), coords=new_coords)

    def copy(self, data=None):
        """
        Create a copy of this UngriddedData object with new data and coordinates
        so that that they can be modified without held references being affected.
        Will call any lazy loading methods in the data and coordinates

        :param ndarray data: Replace the data of the ungridded data copy with provided data

        :return: Copied UngriddedData object
        """
        from copy import deepcopy
        data = data if data is not None else numpy.ma.copy(self.data)  # This will load the data if lazy load
        coords = self.coords().copy()
        return UngriddedData(data=data, metadata=deepcopy(self.metadata), coords=coords)

    @property
    def size(self):
        return self.data.size

    def count(self):
        return self.data.count() if hasattr(self.data, 'count') else self.data.size

    @property
    def history(self):
        return self.metadata.history

    @property
    def x(self):
        return self.coord(axis='X')

    @property
    def y(self):
        return self.coord(axis='Y')

    @property
    def lat(self):
        return self.coord(standard_name='latitude')

    @property
    def lon(self):
        return self.coord(standard_name='longitude')

    @property
    def time(self):
        return self.coord(standard_name='time')

    def hyper_point(self, index):
        """
        :param index: The index in the array to find the point for
        :return: A hyperpoint representing the data at that point
        """
        from cis.data_io.hyperpoint import HyperPoint
        return HyperPoint(self.coord(standard_name='latitude').data.flat[index],
                          self.coord(standard_name='longitude').data.flat[index],
                          self.coord(standard_name='altitude').data.flat[index],
                          self.coord(standard_name='time').data.flat[index],
                          self.coord(standard_name='air_pressure').data.flat[index],
                          self.data.flat[index])

    def as_data_frame(self, copy=True, time_index=True, name=None):
        """
        Convert an UngriddedData object to a Pandas DataFrame.

        :param copy: Create a copy of the data for the new DataFrame? Default is True.
        :return: A Pandas DataFrame representing the data and coordinates. Note that this won't include any metadata.
        """
        df = _coords_as_data_frame(self.coords(), time_index=time_index)
        try:
            df[name or self.name()] = _to_flat_ndarray(self.data, copy)
        except ValueError:
            logging.warn("Copy created of MaskedArray for {} when creating Pandas DataFrame".format(self.name()))
            df[name or self.name()] = _to_flat_ndarray(self.data, True)

        return df

    def coords(self, name_or_coord=None, standard_name=None, long_name=None, attributes=None, axis=None, var_name=None,
               dim_coords=True):
        """
        :return: A list of coordinates in this UngriddedData object fitting the given criteria
        """
        self._post_process()
        return self._coords.get_coords(name_or_coord, standard_name, long_name, attributes, axis, var_name)

    def coord(self, name_or_coord=None, standard_name=None, long_name=None, attributes=None, axis=None, var_name=None):
        """
        :raise: CoordinateNotFoundError
        :return: A single coord given the same arguments as :meth:`coords`.
        """
        return self.coords().get_coord(name_or_coord, standard_name, long_name, attributes, axis, var_name)

    def get_coordinates_points(self):
        """Returns a HyperPointView of the coordinates of points.

        :return: HyperPointView of the coordinates of points
        """
        return UngriddedHyperPointView(self.coords_flattened, None)

    def get_all_points(self):
        """Returns a HyperPointView of the points.

        :return: HyperPointView of all the data points
        """
        return UngriddedHyperPointView(self.coords_flattened, self.data_flattened)

    def get_non_masked_points(self):
        """Returns a HyperPointView for which the default iterator omits masked points.

        :return: HyperPointView of the data points
        """
        return UngriddedHyperPointView(self.coords_flattened, self.data_flattened, non_masked_iteration=True)

    def find_standard_coords(self):
        """Constructs a list of the standard coordinates.
        The standard coordinates are latitude, longitude, altitude, air_pressure and time; they occur in the return
        list in this order.

        :return: list of coordinates or None if coordinate not present
        """
        return self.coords().find_standard_coords()

    @property
    def is_gridded(self):
        """Returns value indicating whether the data/coordinates are gridded.
        """
        return False

    @classmethod
    def from_points_array(cls, hyperpoints):
        """
        Constuctor for building an UngriddedData object from a list of hyper points

        :param hyperpoints: list of HyperPoints
        """
        from cis.data_io.Coord import Coord, CoordList
        from cis.data_io.hyperpoint import HyperPointList
        from cis.time_util import cis_standard_time_unit

        if not isinstance(hyperpoints, HyperPointList):
            hyperpoints = HyperPointList(hyperpoints)

        values = hyperpoints.vals
        latitude = hyperpoints.latitudes
        longitude = hyperpoints.longitudes
        air_pressure = hyperpoints.air_pressures
        altitude = hyperpoints.altitudes
        time = hyperpoints.times

        coord_list = []
        if latitude is not None:
            coord_list.append(Coord(latitude, Metadata(standard_name='latitude', units='degrees north')))
        if longitude is not None:
            coord_list.append(Coord(longitude, Metadata(standard_name='longitude', units='degrees east')))
        if air_pressure is not None:
            coord_list.append(Coord(air_pressure, Metadata(standard_name='air_pressure', units='Pa')))
        if altitude is not None:
            coord_list.append(Coord(altitude, Metadata(standard_name='altitude', units='meters')))
        if time is not None:
            coord_list.append(Coord(time, Metadata(standard_name='time', units=cis_standard_time_unit)))
        coords = CoordList(coord_list)

        return cls(values, Metadata(), coords)

    def summary(self, shorten=False):
        """
        Unicode summary of the UngriddedData with metadata of itself and its coordinates
        """
        summary = 'Ungridded data: {name} / ({units}) \n'.format(name=self.name(), units=self.units)
        if shorten:
            return summary

        summary += '     Shape = {}\n'.format(self.data.shape) + '\n'
        summary += '     Total number of points = {}\n'.format(self.size)
        summary += '     Number of non-masked points = {}\n'.format(self.count())

        summary += str(self.metadata)

        summary += '     Coordinates: \n'
        for c in self.coords():
            summary += '{pad:{width}}{name}\n'.format(pad=' ', width=7, name=c.name())
            c.update_range()
            summary += c.metadata.summary(offset=10)

        return summary

    def __repr__(self):
        return "<cis 'UngriddedData' of %s>" % self.summary(shorten=True)

    def __str__(self):
        # six has a decorator for this bit, but it doesn't do errors='replace'.
        if six.PY3:
            return self.summary()
        else:
            return self.summary().encode(errors='replace')

    def __unicode__(self):
        return self.summary()

    def set_longitude_range(self, range_start):
        """
        Rotates the longitude coordinate array and changes its values by
        360 as necessary to force the values to be within a 360 range starting
        at the specified value.
        :param range_start: starting value of required longitude range
        """
        from cis.utils import fix_longitude_range
        self.coord(standard_name='longitude').data = fix_longitude_range(self.lon.points, range_start)

    def subset(self, **kwargs):
        """
        Subset the data based on the specified constraints. Note that the limits are inclusive.

        The subset region is defined by passing keyword arguments for each dimension to be subset over, each argument
        must be a slice, or have two entries (a maximum and a minimum). Datetime objects can be used to specify upper
        and lower datetime limits, or a single PartialDateTime object can be used to specify a datetime range.

        The keyword keys are used to find the relevant coordinate, they are looked for in order of name, standard_name,
        axis and var_name.

        For example:
            data.subset(x=[0, 80], y=slice(10, 50))

        or:
            data.aggregate(t=PartialDateTime(2008,9))


        A shape keyword can also be supplied as a WKT string or shapely object to subset in lat/lon by an arbitrary
        shape. In this case the lat/lon bounds are taken as the bounding box of the shape.

        :param kwargs: The constraints for each coordinate dimension
        :return CommonData:
        """
        from cis.subsetting.subset import subset, UngriddedSubsetConstraint
        return subset(self, UngriddedSubsetConstraint, **kwargs)

    def aggregate(self, how=None, **kwargs):
        """
        Aggregate the UngriddedData object based on the specified grids. The grid is defined by passing keyword
        arguments for each dimension, each argument must be a slice, or have three entries (a maximum, a minimum and a
        gridstep). The default aggregation method ('moments') returns the mean, standard deviation and number of points
         as separate GriddedData objects.

        Datetime objects can be used to specify upper and lower datetime limits, or a
        single PartialDateTime object can be used to specify a datetime range. The gridstep can be specified as a
        DateTimeDelta object.

        The keyword keys are used to find the relevant coordinate, they are looked for in order of name, standard_name,
        axis and var_name.

        For example:
            data.aggregate(x=[-180, 180, 360], y=slice(-90, 90, 10))

        or:
            data.aggregate(how='mean', t=[PartialDateTime(2008,9), timedelta(days=1))

        :param str how: The kernel to use in the aggregation (moments, mean, min, etc...). Default is moments
        :param kwargs: The grid specifications for each coordinate dimension
        :return GriddedData:
        """
        agg = _aggregate_ungridded(self, how, **kwargs)
        # Return the single item if there's only one (this depends on the kernel used)
        if len(agg) == 1:
            agg = agg[0]
        return agg

    def sampled_from(self, data, how='', kernel=None, missing_data_for_missing_sample=True, fill_value=None,
                     var_name='', var_long_name='', var_units='', **kwargs):
        """
        Collocate the CommonData object with another CommonData object using the specified collocator and kernel

        :param CommonData or CommonDataList data: The data to resample
        :param str how: Collocation method (e.g. lin, nn, bin or box)
        :param str or cis.collocation.col_framework.Kernel kernel:
        :param bool missing_data_for_missing_sample: Should missing values in sample data be ignored for collocation?
        :param float fill_value: Value to use for missing data
        :param str var_name: The output variable name
        :param str var_long_name: The output variable's long name
        :param str var_units: The output variable's units
        :return CommonData: The collocated dataset
        """
        return _ungridded_sampled_from(self, data, how=how, kernel=kernel,
                                       missing_data_for_missing_sample=missing_data_for_missing_sample,
                                       fill_value=fill_value, var_name=var_name, var_long_name=var_long_name,
                                       var_units=var_units, **kwargs)

    def _get_default_plot_type(self, lat_lon=False):
        if lat_lon:
            return 'scatter2d'
        else:
            return 'line'


class UngriddedCoordinates(CommonData):
    """
    Wrapper (adaptor) class for the different types of possible ungridded data.
    """

    def __init__(self, coords):
        """
        Constructor

        :param coords: A list of the associated Coord objects
        """
        from cis.data_io.Coord import CoordList, Coord

        if isinstance(coords, list):
            self._coords = CoordList(coords)
        elif isinstance(coords, CoordList):
            self._coords = coords
        elif isinstance(coords, Coord):
            self._coords = CoordList([coords])
        else:
            raise ValueError("Invalid Coords type")
        self._post_process()

    def _post_process(self):
        """
        Perform a post processing step on lazy loaded Coordinate Data

        :return:
        """
        # Remove any points with missing coordinate values:
        combined_mask = numpy.zeros(self._coords[0].data_flattened.shape, dtype=bool)
        for coord in self._coords:
            combined_mask |= numpy.ma.getmaskarray(coord.data_flattened)
            if coord.data.dtype != 'object':
                combined_mask |= numpy.isnan(coord.data).flatten()
        if combined_mask.any():
            n_points = numpy.count_nonzero(combined_mask)
            logging.warning("Identified {n_points} point(s) which were missing values for some or all coordinates - "
                            "these points have been removed from the data.".format(n_points=n_points))
            for coord in self._coords:
                coord.data = numpy.ma.masked_array(coord.data_flattened, mask=combined_mask).compressed()
                coord.update_shape()
                coord.update_range()

    @property
    def coords_flattened(self):
        all_coords = self.coords().find_standard_coords()
        return [(c.data_flattened if c is not None else None) for c in all_coords]

    @property
    def history(self):
        return "UngriddedCoordinates have no history"

    @property
    def size(self):
        if len(self._coords) > 1:
            return self._coords[0].data.size
        else:
            return 0

    def count(self):
        # There can be no masked coordinate points
        return self.size

    @property
    def x(self):
        return self.coord(axis='X')

    @property
    def y(self):
        return self.coord(axis='Y')

    @property
    def lat(self):
        return self.coord(standard_name='latitude')

    @property
    def lon(self):
        return self.coord(standard_name='longitude')

    @property
    def time(self):
        return self.coord(standard_name='time')

    @property
    def var_name(self):
        return ''

    def hyper_point(self, index):
        """
        :param index: The index in the array to find the point for
        :return: A hyperpoint representing the data at that point
        """
        from cis.data_io.hyperpoint import HyperPoint
        return HyperPoint(self.coord(standard_name='latitude').data.flat[index],
                          self.coord(standard_name='longitude').data.flat[index],
                          self.coord(standard_name='altitude').data.flat[index],
                          self.coord(standard_name='time').data.flat[index],
                          self.coord(standard_name='air_pressure').data.flat[index],
                          None)

    def as_data_frame(self, copy=True, time_index=True, name=None):
        """
        Convert an UngriddedCoordinates object to a Pandas DataFrame.

        :param copy: Create a copy of the data for the new DataFrame? Default is True.
        :return: A Pandas DataFrame representing the data and coordinates. Note that this won't include any metadata.
        """
        return _coords_as_data_frame(self._coords, time_index=time_index)

    def coords(self, name_or_coord=None, standard_name=None, long_name=None, attributes=None, axis=None, var_name=None,
               dim_coords=True):
        """
        :return: A list of coordinates in this UngriddedData object fitting the given criteria
        """
        return self._coords.get_coords(name_or_coord, standard_name, long_name, attributes, axis, var_name)

    def coord(self, name_or_coord=None, standard_name=None, long_name=None, attributes=None, axis=None, var_name=None):
        """
        :raise: CoordinateNotFoundError
        :return: A single coord given the same arguments as :meth:`coords`.

        """
        return self._coords.get_coord(name_or_coord, standard_name, long_name, attributes, axis, var_name)

    def get_coordinates_points(self):
        return UngriddedHyperPointView(self.coords_flattened, None)

    def get_all_points(self):
        """Returns a HyperPointView of the points.

        :return: HyperPointView of all the data points
        """
        return UngriddedHyperPointView(self.coords_flattened, None)

    def get_non_masked_points(self):
        """Returns a HyperPointView for which the default iterator omits masked points.

        :return: HyperPointView of the data points
        """
        return UngriddedHyperPointView(self.coords_flattened, None, non_masked_iteration=True)

    @property
    def is_gridded(self):
        """Returns value indicating whether the data/coordinates are gridded.
        """
        return False

    def set_longitude_range(self, range_start):
        """
        Rotates the longitude coordinate array and changes its values by
        360 as necessary to force the values to be within a 360 range starting
        at the specified value.
        :param range_start: starting value of required longitude range
        """
        from cis.utils import fix_longitude_range
        self.coord(standard_name='longitude').data = fix_longitude_range(self.lon.points, range_start)

    def subset(self, **kwargs):
        raise NotImplementedError("Subset is not available for UngriddedCoordinates objects")

    def collocated_onto(self, sample, how='', kernel=None, **kwargs):
        raise NotImplementedError("UngriddedCoordinates objects cannot be used as sources of data for collocation.")

    def sampled_from(self, data, how='', kernel=None, missing_data_for_missing_sample=False, fill_value=None,
                     var_name='', var_long_name='', var_units='', **kwargs):
        """
        Collocate the CommonData object with another CommonData object using the specified collocator and kernel

        Note - that the default value for missing_data_for_missing_sample is different in this implementation as
        compared to the UngriddedData implementation.

        :param CommonData or CommonDataList data: The data to resample
        :param str how: Collocation method (e.g. lin, nn, bin or box)
        :param str or cis.collocation.col_framework.Kernel kernel:
        :param bool missing_data_for_missing_sample: Should missing values in sample data be ignored for collocation?
        :param float fill_value: Value to use for missing data
        :param str var_name: The output variable name
        :param str var_long_name: The output variable's long name
        :param str var_units: The output variable's units
        :return CommonData: The collocated dataset
        """
        return _ungridded_sampled_from(self, data, how=how, kernel=kernel,
                                       missing_data_for_missing_sample=missing_data_for_missing_sample,
                                       fill_value=fill_value, var_name=var_name, var_long_name=var_long_name,
                                       var_units=var_units, **kwargs)

    def _get_default_plot_type(self, lat_lon=False):
        raise NotImplementedError("UngriddedCoordinates have no default plot type")

    def var_name(self):
        raise NotImplementedError("UngriddedCoordinates have no var name")


class UngriddedDataList(CommonDataList):
    """
    Class which represents multiple UngriddedData objects (e.g. from reading multiple variables)
    """

    def __str__(self):
        return "UngriddedDataList: \n%s" % super(UngriddedDataList, self).__str__()

    @property
    def is_gridded(self):
        """Returns value indicating whether the data/coordinates are gridded.
        """
        return False

    def save_data(self, output_file):
        """
        Save the UngriddedDataList to a file

        :param output_file: output filename
        :return:
        """
        logging.info('Saving data to %s' % output_file)
        # Should only write coordinates out once
        write_coordinates(self[0], output_file)
        for data in self:
            add_data_to_file(data, output_file)

    def get_non_masked_points(self):
        """
        Returns a list containing a HyperPointViews for which the default iterator omits masked points, for each item in
        this UngriddedDataList.

        :return: List of HyperPointViews of the data points
        """
        points_list = []
        for data in self:
            points_list.append(data.get_non_masked_points())
        return points_list

    def coord(self, *args, **kwargs):
        """
        Call :func:`UnGriddedData.coord(*args, **kwargs)` for the first item of data (assumes all data in list has
        same coordinates)

        :param args:
        :param kwargs:
        :return:
        """
        return self[0].coord(*args, **kwargs)

    def copy(self):
        """
        Create a copy of this UngriddedDataList with new data and coordinates
        so that that they can be modified without held references being affected.
        Will call any lazy loading methods in the data and coordinates

        :return: Copied UngriddedData object
        """
        output = UngriddedDataList()
        for variable in self:
            output.append(variable.copy())
        return output

    def as_data_frame(self, copy=True):
        """
        Convert an UngriddedDataList object to a Pandas DataFrame. Note that UngriddedDataList objects are expected to
        share coordinates, so only the coordinates from the first object in the list are used.

        :param copy: Create a copy of the data for the new DataFrame? Default is True.
        :return: A Pandas DataFrame representing the data and coordinates. Note that this won't include any metadata.

        .. note::
            This function will copy your data by default.
            If you have a large array that cannot be copied,
            make sure it is not masked and use copy=False.
        """
        import numpy as np

        df = self[0].as_data_frame(copy=copy)

        for d in self[1:]:
            try:
                data = _to_flat_ndarray(d.data, copy)
            except ValueError:
                logging.warn("Copy created of MaskedArray for {} when creating Pandas DataFrame".format(d.name()))
                data = _to_flat_ndarray(d.data, True)
            df[d.name()] = data

        return df

    def subset(self, **kwargs):
        from cis.subsetting.subset import subset, UngriddedSubsetConstraint
        return subset(self, UngriddedSubsetConstraint, **kwargs)

    def aggregate(self, how='', **kwargs):
        """
        Aggregate the UngriddedDataList object based on the specified grids. The grid is defined by passing keyword
        arguments for each dimension, each argument must be a slice, or have three entries (a maximum, a minimum and a
        gridstep). The default aggregation method ('moments') returns the mean, standard deviation and number of points
         as separate GriddedData objects (for each UngriddedData object in the list).

        Datetime objects can be used to specify upper and lower datetime limits, or a
        single PartialDateTime object can be used to specify a datetime range. The gridstep can be specified as a
        DateTimeDelta object.

        The keyword keys are used to find the relevant coordinate, they are looked for in order of name, standard_name,
        axis and var_name.

        For example:
            data.aggregate(x=[-180, 180, 360], y=slice(-90, 90, 10))

        or:
            data.aggregate(how='mean', t=[PartialDateTime(2008,9), timedelta(days=1))

        :param str how: The kernel to use in the aggregation (moments, mean, min, etc...)
        :param kwargs: The grid specifications for each coordinate dimension
        :return GriddedDataList:
        """
        return _aggregate_ungridded(self, how, **kwargs)


def _coords_as_data_frame(coord_list, copy=True, time_index=True):
    """
    Convert a CoordList object to a Pandas DataFrame.

    :param copy: Create a copy of the data for the new DataFrame? Default is True.
    :return: A Pandas DataFrame representing the data and coordinates. Note that this won't include any metadata.
    """
    import pandas as pd
    from cis.time_util import cis_standard_time_unit
    from cf_units import Unit

    columns = {}
    time = None

    for coord in coord_list.get_coords():
        try:
            data = _to_flat_ndarray(coord.data, copy)
        except ValueError:
            logging.warn("Copy created of MaskedArray for {} when creating Pandas DataFrame".format(coord.name()))
            data = _to_flat_ndarray(coord.data, True)

        if time_index and coord.standard_name == 'time':
            if str(coord.units).lower() == 'datetime object':
                time = data
            elif isinstance(coord.units, Unit):
                time = coord.units.num2date(data)
            else:
                time = cis_standard_time_unit.num2date(data)
        else:
            columns[coord.standard_name] = data

    return pd.DataFrame(columns, index=time)


def _to_flat_ndarray(data, copy=True):
    """
    Convert a (possibly masked) numpy array into its flat equivalent, with or without copying it.

    :param data:
    :param copy:
    :return:
    """
    import numpy as np

    if isinstance(data, np.ma.MaskedArray):
        if not copy:
            raise ValueError("Masked arrays must always be copied.")
        # We need to cast the array to a float so that we can fill the array with NaNs for Pandas (which would do the
        #  same trick itself anyway)
        ndarr = data.astype(np.float64).filled(np.NaN).flatten()
    elif copy:
        ndarr = data.flatten()
    else:
        ndarr = data.ravel()

    return ndarr


def _ungridded_sampled_from(sample, data, how='', kernel=None, missing_data_for_missing_sample=True, fill_value=None,
                            var_name='', var_long_name='', var_units='', **kwargs):
    """
    Collocate the CommonData object with another CommonData object using the specified collocator and kernel

    :param CommonData or CommonDataList data: The data to resample
    :param str how: Collocation method (e.g. lin, nn, bin or box)
    :param str or cis.collocation.col_framework.Kernel kernel:
    :param bool missing_data_for_missing_sample: Should missing values in sample data be ignored for collocation?
    :param float fill_value: Value to use for missing data
    :param str var_name: The output variable name
    :param str var_long_name: The output variable's long name
    :param str var_units: The output variable's units
    :return CommonData: The collocated dataset
    """
    from cis.collocation import col_implementations as ci
    from cis.data_io.gridded_data import GriddedData, GriddedDataList
    from cis.collocation.col import collocate, get_kernel

    if isinstance(data, UngriddedData) or isinstance(data, UngriddedDataList):
        col = ci.GeneralUngriddedCollocator(fill_value=fill_value, var_name=var_name, var_long_name=var_long_name,
                                            var_units=var_units,
                                            missing_data_for_missing_sample=missing_data_for_missing_sample)

        # Box is the default, and only option for ungridded -> ungridded collocation
        if how not in ['', 'box']:
            raise ValueError("Invalid method specified for ungridded -> ungridded collocation: " + how)
        con = ci.SepConstraintKdtree(**kwargs)
        # We can have any kernel, default to moments
        kernel = get_kernel(kernel)
    elif isinstance(data, GriddedData) or isinstance(data, GriddedDataList):
        col = ci.GriddedUngriddedCollocator(fill_value=fill_value, var_name=var_name, var_long_name=var_long_name,
                                            var_units=var_units,
                                            missing_data_for_missing_sample=missing_data_for_missing_sample)
        con = None
        kernel = 'lin'
    else:
        raise ValueError("Invalid argument, data must be either GriddedData or UngriddedData")

    return collocate(data, sample, col, con, kernel)


def _aggregate_ungridded(data, how, **kwargs):
    """
    Aggregate an UngriddedData or UngriddedDataList based on the specified grids
    :param UngriddedData or UngriddedDataList data: The data object to aggregate
    :param cis.collocation.col_framework.Kernel kernel: The kernel to use in the aggregation
    :param kwargs: The grid specifications for each coordinate dimension
    :return:
    """
    from cis.aggregation.ungridded_aggregator import UngriddedAggregator
    from cis.collocation.col import get_kernel
    from cis.time_util import PartialDateTime
    from datetime import datetime, timedelta
    from cis import __version__

    kernel = get_kernel(how)
    grid_spec = {}
    for dim_name, grid in kwargs.items():
        c = data._get_coord(dim_name)
        if all(hasattr(grid, att) for att in ('start', 'stop', 'step')):
            g = grid
        elif len(grid) == 2 and isinstance(grid[0], PartialDateTime):
            g = slice(grid[0].min(), grid[0].max(), grid[1])
        elif len(grid) == 3:
            g = slice(grid[0], grid[1], grid[2])
        else:
            raise ValueError("Invalid subset arguments: {}".format(grid))

        # Fill in defaults
        grid_start = g.start if g.start is not None else c.points.min()
        if isinstance(grid_start, datetime):
            grid_start = c.units.date2num(grid_start)

        grid_end = g.stop if g.stop is not None else c.points.max()
        if isinstance(grid_end, datetime):
            grid_end = c.units.date2num(grid_end)

        if g.step is None:
            raise ValueError("Grid step must not be None")
        else:
            grid_step = g.step

        if isinstance(grid_step, timedelta):
            # Standard time is days since, so turn this into a fractional number of days
            grid_step = grid_step.total_seconds() / (24*60*60)

        grid_spec[c.name()] = slice(grid_start, grid_end, grid_step)

    # We have to make the history before doing the aggregation as the grid dims get popped-off during the operation
    history = "Aggregated using CIS version " + __version__ + \
              "\n variables: " + str(getattr(data, "var_name", "Unknown")) + \
              "\n from files: " + str(getattr(data, "filenames", "Unknown")) + \
              "\n using new grid: " + str(grid_spec) + \
              "\n with kernel: " + str(kernel) + "."

    aggregator = UngriddedAggregator(grid_spec)
    data = aggregator.aggregate(data, kernel)

    data.add_history(history)

    return data



from setuptools.command.test import test as TestCommand
import multiprocessing


def run(test_set='cis/test/unit', n_processors=1, stop=False, debug=False):
    import nose
    import logging
    import sys

    if debug:
        logging.basicConfig(level=logging.DEBUG,
                            format="%(asctime)s - %(levelname)s - %(module)s : %(lineno)d - %(message)s",
                            stream=sys.stdout)

    args = ['', test_set, '--processes=%s' % n_processors, '--verbosity=2', '--process-timeout=60']

    if stop:
        args.append('--stop')

    nose.run_exit(argv=args)


class nose_test(TestCommand):
    """
    Command to run unit tests
    """
    description = "Run CIS tests. By default this will run all of the unit tests. Optionally the integration tests can"\
                  " be run instead."
    user_options = [('integration-tests', 'i', 'Run the integration tests.'),
                    ('stop', 'x', 'Stop running tests after the first error or failure.'),
                    ('num-processors=', 'p', 'The number of processors used for running the tests.'),
                    ('debug', 'd', 'Send debug level logging information to stdout.')]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.integration_tests = False
        self.stop = False
        self.num_processors = None
        self.debug = False

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
        if self.integration_tests:
            self.test_set = 'cis/test/integration'
        else:
            self.test_set = 'cis/test/unit'

        if self.num_processors is None:
            self.num_processors = multiprocessing.cpu_count() - 1
        else:
            self.num_processors = int(self.num_processors)

    def run_tests(self):
        run(self.test_set, self.num_processors, self.stop, self.debug)

        # nose.run_exit(argv=['nosetests',os.path.join(os.path.dirname(__file__), 'unit')])

